/* $Id: options.c,v 1.73 2010/10/07 19:18:29 lamonte Exp $
 *
 * options.c
 *
 * Additional acknowledgements, news, and updates to the Qodem Project
 * are available at http://qodem.sourceforge.net .
 *
 * This module is licensed under the GNU General Public License Version 2.
 * Please see the file "COPYING" in this directory for more information
 * about the GNU General Public License Version 2.
 *
 *     Copyright (C) 2003  Kevin Lamonte
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

#include "common.h"

#include <stdio.h>			/* fopen(), fgets(), fclose() */
#include <stdlib.h>			/* getenv() */
#include <string.h>			/* strstr(), strncpy(), memset() */
#include <unistd.h>			/* access() */
#include <ctype.h>			/* isspace(), tolower() */
#include <sys/stat.h>			/* S_IRUSR | S_IWUSR | S_IXUSR */
#include <sys/types.h>			/* mkdir() */
#include <errno.h>
#include "keyboard.h"
#include "translate.h"
#include "modem.h"
#include "qodem.h"
#include "input.h"
#include "states.h"
#include "status.h"
#include "options.h"

/* Options structure.  I think this is the most useless comment so far. :-) */
struct option_struct {
	Q_OPTION option;
	char * value;
	char * name;
	char * default_value;
	char * comment;
};

/* Options list */
static struct option_struct options[] = {
	{Q_OPTION_WORKING_DIR, NULL, "working_dir", "$HOME/qodem", ""
"### The default working directory.  If this directory does not exist\n"
"### on startup, the user will be prompted whether or not to create it.\n"
"### The $HOME environment variable will be substituted if specified."},
	{Q_OPTION_ISO8859_LANG, NULL, "iso8859_lang", "C", ""
"### The LANG environment variable to specify for the remote\n"
"### connection for non-Unicode emulations."},
	{Q_OPTION_UTF8_LANG, NULL, "utf8_lang", "en_US.UTF-8", ""
"### The LANG environment variable to specify for the remote\n"
"### connection for LINUX UTF-8 and XTERM UTF-8 emulations."},
	{Q_OPTION_SOUNDS_ENABLED, NULL, "sounds", "true", ""
"### Whether or not to support sounds.  This overrides ANSI music.  Value\n"
"### is 'true' or 'false'."},
	{Q_OPTION_ANSI_MUSIC, NULL, "ansi_music", "true", ""
"### Whether or not ANSI music is enabled on startup.  Value is 'true'\n"
"### or 'false'."},
	{Q_OPTION_ANSI_ANIMATE, NULL, "ansi_animate", "false", ""
"### Whether or not ANSI should update the screen quickly to support animation.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means that ANSI emulation will update the screen much more often,\n"
"### resulting in better animation sequences at a high performance penalty.\n"
"### 'false' means buffer ANSI output like all other emulations."},
	{Q_OPTION_80_COLUMNS, NULL, "80_columns", "true", ""
"### Whether or not ANSI, AVATAR, and TTY emulations assume 80 columns.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means lines will wrap properly (if line wrap is enabled) at\n"
"### column 80.  This is often needed when connecting to text-based BBSes\n"
"### with classic ANSI art screens."},
	{Q_OPTION_START_PHONEBOOK, NULL, "start_in_phonebook", "true", ""
"### Whether to startup in the phonebook.  Value is 'true' or\n"
"### 'false'."},
	{Q_OPTION_CONNECT_DOORWAY, NULL, "doorway_mode_on_connect", "off", ""
"### Whether to automatically switch to DOORWAY or MIXED mode after connecting.\n"
"### Value is 'doorway', 'mixed', or 'off'."},
	{Q_OPTION_DOORWAY_MIXED_KEYS, NULL, "doorway_mixed_mode_commands", "D P T Y Z / PgUp PgDn", ""
"### A space-separated list of command keys that will be honored when in MIXED\n"
"### doorway mode.  Each key is one of the Alt-key combos on the Alt-Z Command menu,\n"
"### except for 'PgUp' and 'PgDn'.  Listing 'PgUp' or 'PgDn' here means to allow\n"
"### the unmodified 'PgUp' and 'PgDn' keys to go to the remote side but still honor\n"
"### ALT- and CTRL- 'PgUp' and 'PgDn'.\n"
"### The default commands to honor are:\n"
"###     Alt-D Phonebook\n"
"###     Alt-P Capture\n"
"###     Alt-T Screen Dump\n"
"###     Alt-Y COM Parameters\n"
"###     Alt-Z Menu\n"
"###     Alt-/ Scrollback\n"
"###     Alt-PgUp or Ctrl-PgUp Upload Files\n"
"###     Alt-PgDn or Ctrl-PgDn Download Files"
	},
	{Q_OPTION_SHELL, NULL, "shell", "/bin/bash --login", ""
"### The OS shell program.  Examples: /bin/bash /bin/tcsh my_shell"},
	{Q_OPTION_SSH, NULL, "ssh", "ssh -e none $REMOTEHOST -p $REMOTEPORT", ""
"### The ssh connection program.  Examples: /bin/ssh /usr/local/bin/ssh2\n"
"###\n"
"### The default value includes the -e none option to disable the escape\n"
"### character.  This arguments works for the Debian Linux OpenSSH 3.8\n"
"### client.  You may have to change it for your client.\n"
"###\n"
"### $REMOTEHOST will be replaced with the phonebook address,\n"
"### $REMOTEPORT will be replaced with the phonebook port."},
	{Q_OPTION_SSH_USER, NULL, "ssh", "ssh -e none -l $USERNAME -p $REMOTEPORT $REMOTEHOST", ""
"### The ssh connection program when the phonebook username is set.\n"
"###\n"
"### The default value includes the -e none option to disable the escape\n"
"### character.  This arguments works for the Debian Linux OpenSSH 3.8\n"
"### client.  You may have to change it for your client.\n"
"###\n"
"### $USERNAME will be replaced with the phonebook username, $REMOTEHOST\n"
"### will be replaced with the phonebook address, and $REMOTEPORT\n"
"### will be replaced with the phonebook port."},
{Q_OPTION_RSH, NULL, "rsh", "rsh $REMOTEHOST", ""
"### The rsh connection program.  Examples: /bin/rsh /usr/local/bin/rsh\n"
"###\n"
"### $REMOTEHOST will be replaced with the phonebook address."},
	{Q_OPTION_RSH_USER, NULL, "rsh", "rsh -l $USERNAME $REMOTEHOST", ""
"### The rsh connection program when the phonebook username is set.\n"
"###\n"
"### $USERNAME will be replaced with the phonebook username and $REMOTEHOST\n"
"### will be replaced with the phonebook address."},
	{Q_OPTION_RLOGIN, NULL, "rlogin", "rlogin $REMOTEHOST", ""
"### The rlogin connection program.  Examples: /bin/rlogin\n"
"### /usr/local/bin/rlogin\n"
"###\n"
"### $REMOTEHOST will be replaced with the phonebook address."},
	{Q_OPTION_RLOGIN_USER, NULL, "rlogin_user", "rlogin -l $USERNAME $REMOTEHOST", ""
"### The rlogin connection program to use when the phonebook username is set.\n"
"###\n"
"### $USERNAME will be replaced with the phonebook username and $REMOTEHOST\n"
"### will be replaced with the phonebook address."},
	{Q_OPTION_TELNET, NULL, "telnet", "telnet -E -8 $REMOTEHOST $REMOTEPORT", ""
"### The telnet connection program.  Examples: /bin/telnet\n"
"### /usr/local/bin/telnet\n"
"###\n"
"### The default value includes the -E option to disable the escape\n"
"### character and the -8 option to negotiate an 8-bit connection.\n"
"### These arguments work for the Debian Linux telnet client.  You may\n"
"### have to change it for more traditional Unix-like operating system\n"
"### clients.\n"
"###\n"
"### $REMOTEHOST will be replaced with the phonebook address,\n"
"### $REMOTEPORT will be replaced with the phonebook port."},
	/*
	{Q_OPTION_TN3270, NULL, "tn3270", "c3270 $REMOTEHOST", ""
"### The tn3270 connection program.  Examples: /bin/tn3270\n"
"### /usr/local/bin/c3270\n"
"###\n"
"### $USERNAME will be replaced with the phonebook username and $REMOTEHOST\n"
"### will be replaced with the phonebook address."},
	 */
	{Q_OPTION_EDITOR, NULL, "editor", "vi", ""
"### The editor program.  The $EDITOR environment variable will be\n"
"### substituted if specified."},
	{Q_OPTION_CAPTURE, NULL, "capture_enabled", "false", ""
"### Whether or not capture is enabled on startup.  Value is\n"
"### 'true' or 'false'."},
	{Q_OPTION_CAPTURE_FILE, NULL, "capture_file", "capture.txt", ""
"### The default capture file name.  When enabled, all transmitted and\n"
"### received bytes (minus color) are appended to this file.  This file\n"
"### is stored in the working directory if a relative path is specified."},
	{Q_OPTION_CAPTURE_TYPE, NULL, "capture_type", "normal", ""
"### The default capture format.  Value is 'normal', 'raw', or 'html'."},
	{Q_OPTION_LOG, NULL, "log_enabled", "false", ""
"### Whether or not session logging is enabled on startup.  Value is\n"
"### 'true' or 'false'."},
	{Q_OPTION_LOG_FILE, NULL, "log_file", "session_log.txt", ""
"### The default session log file name.  When enabled, an entry is appended\n"
"### to this file for one of the following events:\n"
"###     connect\n"
"###     disconnect\n"
"###     program start\n"
"###     program exit\n"
"###     file upload\n"
"###     file download\n"
"###     OS shell\n"
"###     Scripted timestamp message\n"
"### This file is stored in the working directory if a relative path is\n"
"### specified."},
	{Q_OPTION_DOWNLOAD_DIR, NULL, "download_dir", "$HOME/qodem", ""
"### The default directory to store downloaded files.  The $HOME\n"
"### environment variable will be substituted if specified."},
	{Q_OPTION_UPLOAD_DIR, NULL, "upload_dir", "$HOME/qodem", ""
"### The default directory to look for files to upload.  The $HOME\n"
"### environment variable will be substituted if specified."},
	{Q_OPTION_BATCH_ENTRY_FILE, NULL, "bew_file", "$HOME/qodem/batch_upload.txt", ""
"### Where to store the Batch Entry Window entries."},
	{Q_OPTION_MAIL_READER, NULL, "mail_reader", "mm", ""
"### The QWK/SOUP/etc. mail reader program.  Default is multimail (mm)"},
	{Q_OPTION_IDLE_TIMEOUT, NULL, "idle_timeout", "0", ""
"### The number of idle seconds to wait before automatically closing\n"
"### the connection.  A value of 0 means never disconnect."},
	{Q_OPTION_KEEPALIVE_TIMEOUT, NULL, "keepalive_timeout", "0", ""
"### The number of idle seconds to wait before automatically sending\n"
"### the keepalive bytes.  A value of 0 disables the keepalive feature."},
	{Q_OPTION_KEEPALIVE_BYTES, NULL, "keepalive_bytes", "\\x00", ""
"### The bytes to every 'keepalive_timeout' seconds.  Use C-style\n"
"### hex notation with 2 hex digits to embed raw bytes, e.g. '\\x00' to\n"
"### mean ASCII NUL, '\\x32' is converted to '2', etc.\n"
"###\n"
"### The maximum string size is 128 bytes."},
	{Q_OPTION_SCREENSAVER_TIMEOUT, NULL, "screensaver_timeout", "0", ""
"### The number of idle seconds to wait before automatically locking\n"
"### the screen.  A value of 0 means never lock the screen."},
	{Q_OPTION_SCREENSAVER_PASSWORD, NULL, "screensaver_password", "password", ""
"### The password required to unlock the screen when the screen saver\n"
"### is active."},
	{Q_OPTION_VT52_COLOR, NULL, "vt52_ansi_color", "true", ""
"### Whether or not ANSI.SYS-style color selection commands will be "
"supported\n"
"### with the VT52 emulation.  Value is 'true' or 'false'.\n"
"###\n"
"### Real VT52 applications are in black and white only.  However, some\n"
"### host application send color selection commands despite the fact the\n"
"### VT52 terminfo/terminfo entry lacks these codes.  ('ls' is one notable\n"
"### example.)  If this value is set to true the VT52 emulator will honor\n"
"### the color selection codes.  If this value is false the VT52 emulator\n"
"### will show the broken escape codes on the screen as (presumably) a real\n"
"### VT52 would do."},
	{Q_OPTION_VT100_COLOR, NULL, "vt100_ansi_color", "true", ""
"### Whether or not ANSI.SYS-style color selection commands will be\n"
"### supported with the VT100, VT102, and VT220 emulations.  Value is\n"
"### 'true' or 'false'.\n"
"###\n"
"### Real VT100, VT102, and VT220 applications are in black and white\n"
"### only.  However, some host applications send color selection commands\n"
"### despite the fact the termcap/terminfo entry lacks these codes.\n"
"### If this value is set to true the VT100, VT102, and VT220 emulation\n"
"### will honor the color selection codes.  If this value is false the\n"
"### color selection codes will be quietly consumed, as a real VT100-ish\n"
"### terminal would do."},
	{Q_OPTION_AVATAR_COLOR, NULL, "avatar_ansi_color", "true", ""
"### Whether or not ANSI.SYS-style color selection commands will be\n"
"### supported with the AVATAR emulation.  Value is 'true' or 'false'.\n"
"###\n"
"### Avatar emulation has its own color selection command, but some\n"
"### programs (like 'ls') send it ANSY.SYS-style color commands\n"
"### instead.  If this value is set to true the AVATAR emulation will\n"
"### honor the ANSI.SYS-style color selection codes.  If this value is\n"
"### false the color selection codes will be visible in the output, as a\n"
"### real Avatar emulator would do."},
	{Q_OPTION_ASCII_UPLOAD_USE_TRANSLATE_TABLE, NULL, "ascii_upload_use_xlate_table", "true", ""
"### Whether or not the ASCII translate table function should be used\n"
"### during ASCII file uploads.  Value is 'true' or 'false'.\n"
"###\n"
"### When true, outgoing bytes will first be translated according to the\n"
"### table and then sent to the remote system."},
	{Q_OPTION_ASCII_UPLOAD_CR_POLICY, NULL, "ascii_upload_cr_policy", "none", ""
"### How to handle outgoing carriage-return characters (0x0D)\n"
"### during ASCII file uploads.  Value is 'none', 'strip', or 'add'.\n"
"###\n"
"### 'none' means do nothing to change the bytes sent.\n"
"### 'strip' means remove carriage-returns while sending the file.\n"
"### 'add' means add a linefeed character (0x0A) after each carriage-"
"return\n"
"### while sending the file."},
	{Q_OPTION_ASCII_UPLOAD_LF_POLICY, NULL, "ascii_upload_lf_policy", "none", ""
"### How to handle outgoing linefeed characters (0x0A) during ASCII file\n"
"### uploads.  Value is 'none', 'strip', or 'add'.\n"
"###\n"
"### 'none' means do nothing to change the bytes sent.\n"
"### 'strip' means remove linefeeds while sending the file.\n"
"### 'add' means add a carriage-return character (0x0D) before each "
"linefeed\n"
"### while sending the file."},
	{Q_OPTION_ASCII_DOWNLOAD_USE_TRANSLATE_TABLE, NULL, "ascii_download_use_xlate_table", "true", ""
"### Whether or not the ASCII translate table function should be used\n"
"### during ASCII file downloads.  Value is 'true' or 'false'.\n"
"###\n"
"### When true, incoming bytes will be translated according to the table\n"
"### before being saved to file."},
	{Q_OPTION_ASCII_DOWNLOAD_CR_POLICY, NULL, "ascii_download_cr_policy", "none", ""
"### How to handle incoming carriage-return characters (0x0D)\n"
"### during ASCII file downloads.  Value is 'none', 'strip', or 'add'.\n"
"###\n"
"### 'none' means do nothing to change the bytes saved.\n"
"### 'strip' means remove carriage-returns while saving the file.\n"
"### 'add' means add a linefeed character (0x0A) after each carriage-"
"return\n"
"### while saving the file."},
	{Q_OPTION_ASCII_DOWNLOAD_LF_POLICY, NULL, "ascii_download_lf_policy", "none", ""
"### How to handle incoming linefeed characters (0x0A) during ASCII file\n"
"### downloads.  Value is 'none', 'strip', or 'add'.\n"
"###\n"
"### 'none' means do nothing to change the bytes saved.\n"
"### 'strip' means remove linefeeds while saving the file.\n"
"### 'add' means add a carriage-return character (0x0D) before each "
"linefeed\n"
"### while saving the file."},
	{Q_OPTION_ZMODEM_AUTOSTART, NULL, "zmodem_autostart", "true", ""
"### Whether or not Zmodem autostart should be used.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Zmodem autostart will be enabled.\n"
"### 'false' means Zmodem autostart will not be enabled."},
	{Q_OPTION_ZMODEM_ZCHALLENGE, NULL, "zmodem_zchallenge", "false", ""
"### Whether or not Zmodem will issue ZCHALLENGE at the beginning.\n"
"### of each transfer.  ZCHALLENGE was meant to improve security\n"
"### but some Zmodem clients do not support it.  Its security\n"
"### benefits are dubious.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Zmodem will issue a ZCHALLENGE.\n"
"### 'false' means Zmodem will not issue a ZCHALLENGE."},
	{Q_OPTION_ZMODEM_ESCAPE_CTRL, NULL, "zmodem_escape_control_chars", "false", ""
"### Whether or not Zmodem should escape control characters by default.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Zmodem will escape control characters, which will\n"
"### make file transfers slower but may be necessary for Zmodem to\n"
"### work at all over the link.\n"
"### 'false' means Zmodem will not escape control characters.\n"
"### \n"
"### In both cases, Zmodem will honor the encoding requested at the\n"
"### other end."},
	{Q_OPTION_KERMIT_AUTOSTART, NULL, "kermit_autostart", "true", ""
"### Whether or not Kermit autostart should be enabled by default.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Kermit autostart will be enabled on startup.\n"
"### 'false' means Kermit autostart will not be enabled on startup."},
	{Q_OPTION_KERMIT_ROBUST_FILENAME, NULL, "kermit_robust_filename", "false", ""
"### Whether or not Kermit should use common form filenames.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Kermit uploads will convert filenames to uppercase,\n"
"### remove all but one period, and remove many punctuation characters.\n"
"### 'false' means Kermit uplods will use the literal filename."},
	{Q_OPTION_KERMIT_STREAMING, NULL, "kermit_streaming", "true", ""
"### Whether or not Kermit should use streaming (sending all file data\n"
"### packets continuously without waiting for ACKs).\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Kermit will use streaming, resulting in a significant\n"
"### performance improvement in most cases, especially over TCP links.\n"
"### 'false' means Kermit will not use streaming."},
	{Q_OPTION_KERMIT_UPLOADS_FORCE_BINARY, NULL, "kermit_uploads_force_binary", "true", ""
"### Whether or not Kermit uploads will transfer files as 8-bit binary files.\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Kermit uploads will transfer all files (including text\n"
"### files) in binary.\n"
"### 'false' means Kermit will convert text files to CRLF format, but\n"
"### leave binary files as-is.  Note that Qodem's kermit checks the first\n"
"### 1024 bytes of the file, and if it sees only ASCII characters assumes\n"
"### the file is text; this heuristic might occasionally mis-identify files."},
	{Q_OPTION_KERMIT_DOWNLOADS_CONVERT_TEXT, NULL, "kermit_downloads_convert_text", "false", ""
"### Whether or not Kermit downloads will convert text files to the local\n"
"### end-of-line convention (e.g. CRLF -> LF).\n"
"### Value is 'true' or 'false'.\n"
"###\n"
"### 'true' means Kermit downloads will convert CRLF to LF.\n"
"### 'false' means Kermit will leave text files in the format sent, usually CRLF."},
	{Q_OPTION_DIAL_BETWEEN_TIME, NULL, "dial_between_time", "5", ""
"### How many seconds to wait after a busy signal before dialing\n"
"### the next number."},
	{Q_OPTION_MUSIC_CONNECT, NULL, "music_on_connect", "none", ""
"### If sounds are enabled, the music sequence to play after\n"
"### successfully connected.  The string is the same format used\n"
"### by the GWBASIC PLAY statement (ANSI Music), or 'none'"},
	{Q_OPTION_MUSIC_CONNECT_MODEM, NULL, "music_on_modem_connect", "MN L16 T120 O5 AB>CAB>CAB>C", ""
"### If sounds are enabled, the music sequence to play after\n"
"### successfully connected via modem.  The string is the same format\n"
"### used by the GWBASIC PLAY statement (ANSI Music), or 'none'"},
	{Q_OPTION_MUSIC_UPLOAD, NULL, "music_on_upload_complete", "MS L8 T120 O5 EEEEE", ""
"### If sounds are enabled, the music sequence to play after\n"
"### a successful upload.  The string is the same format used\n"
"### by the GWBASIC PLAY statement (ANSI Music)"},
	{Q_OPTION_MUSIC_DOWNLOAD, NULL, "music_on_download_complete", "MS L8 T120 O5 CCCCC", ""
"### If sounds are enabled, the music sequence to play after\n"
"### a successful download.  The string is the same format used\n"
"### by the GWBASIC PLAY statement (ANSI Music)"},
	{Q_OPTION_NULL, NULL, NULL, NULL}
};

/*
 * substitute_string - replace all instances of "pattern" in "original"
 *                     with "new_string", returning a newly-malloc'd
 *                     string.
 */
char * substitute_string(const char * original, const char * pattern, char * new_string) {
	int i;
	/* Cast to avoid compiler warning */
	char * current_place = (char *)original;
	char * old_place;
	char * return_string;
	int return_string_place;

	/* Count all occurrences */
	i=0;
	while (current_place[0] != '\0') {
		current_place = strstr(current_place, pattern);
		if (current_place == NULL) {
			break;
		} else {
			i++;
			current_place++;
		}
	}

	/* i contains the # of occurrences of pattern. */
	return_string = (char *)Xmalloc(1 + strlen(original) + (i * (strlen(new_string) - strlen(pattern))), __FILE__, __LINE__);
	memset(return_string, 0, 1 + strlen(original) + (i * (strlen(new_string) - strlen(pattern))));
	/* Cast to avoid compiler warning */
	current_place = (char *)original;
	old_place = current_place;
	return_string_place = 0;
	while (i>0) {
		current_place = strstr(old_place, pattern);
		/* Copy part before pattern */
		strncpy(return_string + return_string_place, old_place, current_place - old_place);
		return_string_place += current_place - old_place;
		old_place = current_place;
		/* Copy new string */
		strncpy(return_string + return_string_place, new_string, strlen(new_string));
		return_string_place += strlen(new_string);
		old_place += strlen(pattern);
		current_place = old_place;
		i--;
	}
	/* Copy terminating non-pattern part */
	strncpy(return_string + return_string_place, current_place, original + strlen(original) - current_place);

	return return_string;
} /* ---------------------------------------------------------------------- */

/*
 * substitute_wcs - replace all instances of "pattern" in "original"
 *                  with "new_string", returning a newly-malloc'd
 *                  string.
 */
wchar_t * substitute_wcs(const wchar_t * original, const wchar_t * pattern, wchar_t * new_string) {
	int i;
	/* Cast to avoid compiler warning */
	wchar_t * current_place = (wchar_t *)original;
	wchar_t * old_place;
	wchar_t * return_string;
	int return_string_place;

	/* Count all occurrences */
	i=0;
	while (current_place[0] != L'\0') {
		current_place = wcsstr(current_place, pattern);
		if (current_place == NULL) {
			break;
		} else {
			i++;
			current_place++;
		}
	}

	/* i contains the # of occurrences of pattern. */
	return_string = (wchar_t *)Xmalloc((sizeof (wchar_t)) * (1 + wcslen(original) + (i * (wcslen(new_string) - wcslen(pattern)))), __FILE__, __LINE__);
	wmemset(return_string, 0, 1 + wcslen(original) + (i * (wcslen(new_string) - wcslen(pattern))));
	/* Cast to avoid compiler warning */
	current_place = (wchar_t *)original;
	old_place = current_place;
	return_string_place = 0;
	while (i>0) {
		current_place = wcsstr(old_place, pattern);
		/* Copy part before pattern */
		wcsncpy(return_string + return_string_place, old_place, current_place - old_place);
		return_string_place += current_place - old_place;
		old_place = current_place;
		/* Copy new string */
		wcsncpy(return_string + return_string_place, new_string, wcslen(new_string));
		return_string_place += wcslen(new_string);
		old_place += wcslen(pattern);
		current_place = old_place;
		i--;
	}
	/* Copy terminating non-pattern part */
	wcsncpy(return_string + return_string_place, current_place, original + wcslen(original) - current_place);

	return return_string;
} /* ---------------------------------------------------------------------- */

/*
 * substitute_wcs_half - replace all instances of "pattern" in "original"
 *                       with "new_string", returning a newly-malloc'd
 *                       string.
 */
char * substitute_wcs_half(const char * original, const char * pattern, wchar_t * new_string) {
	int i;
	/* Cast to avoid compiler warning */
	char * current_place = (char *)original;
	char * old_place;
	char * return_string;
	int return_string_place;
	char new_str[256];

	/* Convert to UTF-8 */
	memset(new_str, 0, sizeof(new_str));
	wcstombs(new_str, new_string, wcslen(new_string));

	/* Count all occurrences */
	i=0;
	while (current_place[0] != '\0') {
		current_place = strstr(current_place, pattern);
		if (current_place == NULL) {
			break;
		} else {
			i++;
			current_place++;
		}
	}

	/* i contains the # of occurrences of pattern. */
	return_string = (char *)Xmalloc(1 + strlen(original) + (i * (strlen(new_str) - strlen(pattern))), __FILE__, __LINE__);
	memset(return_string, 0, 1 + strlen(original) + (i * (strlen(new_str) - strlen(pattern))));
	/* Cast to avoid compiler warning */
	current_place = (char *)original;
	old_place = current_place;
	return_string_place = 0;
	while (i>0) {
		current_place = strstr(old_place, pattern);
		/* Copy part before pattern */
		strncpy(return_string + return_string_place, old_place, current_place - old_place);
		return_string_place += current_place - old_place;
		old_place = current_place;
		/* Copy new string */
		strncpy(return_string + return_string_place, new_str, strlen(new_str));
		return_string_place += strlen(new_str);
		old_place += strlen(pattern);
		current_place = old_place;
		i--;
	}
	/* Copy terminating non-pattern part */
	strncpy(return_string + return_string_place, current_place, original + strlen(original) - current_place);

	return return_string;
} /* ---------------------------------------------------------------------- */

/*
 * Retrieve an option value.
 */
char * get_option(const Q_OPTION option) {
	struct option_struct * current_option = options;
	while (current_option->option != Q_OPTION_NULL) {
		if (current_option->option == option) {
			return current_option->value;
		}
		current_option++;
	}
	return NULL;
} /* ---------------------------------------------------------------------- */

/*
 * Save options to options file
 */
static Q_BOOL save_options(const char * filename) {
	struct option_struct * current_option;
	FILE * file;
	int rc;

	file = fopen(filename, "w");
	if (file == NULL) {
		fprintf(stderr, _("Error opening file \"%s\" for writing: %s"), filename, strerror(errno));
		return Q_FALSE;
	}

	current_option = options;
	while (current_option->option != Q_OPTION_NULL) {
		rc = fwrite(current_option->comment, strlen(current_option->comment), 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		rc = fwrite("\n###", 4, 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		rc = fprintf(file, _("\n### Default value: "));
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		rc = fwrite(current_option->default_value, strlen(current_option->default_value), 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		if (strncmp(current_option->value, current_option->default_value, strlen(current_option->default_value)) == 0) {
			rc = fwrite("\n### ", 5, 1, file);
			if (rc == -1) {
				fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
				return Q_FALSE;
			}
		} else {
			rc = fwrite("\n", 1, 1, file);
			if (rc == -1) {
				fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
				return Q_FALSE;
			}
		}
		rc = fwrite(current_option->name, strlen(current_option->name), 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		rc = fwrite(" = ", 3, 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		rc = fwrite(current_option->value, strlen(current_option->value), 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}
		rc = fwrite("\n\n\n", 3, 1, file);
		if (rc == -1) {
			fprintf(stderr, _("Error writing to file \"%s\": %s"), filename, strerror(errno));
			return Q_FALSE;
		}

		current_option++;
	}

	fclose(file);

	return Q_TRUE;
} /* ---------------------------------------------------------------------- */

/*
 * Set option.
 */
static void set_option(struct option_struct * option, char * value) {
	char * new_value = value;
	while (((isspace(new_value[0])) || (new_value[0] == '=')) && (strlen(new_value) > 0)) {
		new_value++;
	}
	new_value = Xstrdup(new_value, __FILE__, __LINE__);;
	if (option->value != NULL) {
		Xfree(option->value, __FILE__, __LINE__);
	}
	option->value = new_value;
} /* ---------------------------------------------------------------------- */

/*
 * Check option.
 */
static void check_option(struct option_struct * option) {
	char * env_string;
	char * new_value;

	/*
	 * Set a default value to empty string so our calls to substitute_string() can
	 * have an Xfree() call.
	 */
	if (option->value == NULL) {
		option->value = Xstrdup("", __FILE__, __LINE__);
	}

	switch (option->option) {
		case Q_OPTION_WORKING_DIR:
		case Q_OPTION_BATCH_ENTRY_FILE:
		case Q_OPTION_UPLOAD_DIR:
		case Q_OPTION_DOWNLOAD_DIR:
			env_string = getenv("HOME");
			if (env_string == NULL) {
				env_string = "";
			}
			/* Sustitute for $HOME */
			new_value = substitute_string(option->value, "$HOME", env_string);
			Xfree(option->value, __FILE__, __LINE__);
			option->value = new_value;
			break;
		case Q_OPTION_EDITOR:
			env_string = getenv("EDITOR");
			if (env_string == NULL) {
				env_string = "";
			}
			/* Sustitute for $EDITOR */
			new_value = substitute_string(option->value, "$EDITOR", env_string);
			Xfree(option->value, __FILE__, __LINE__);
			option->value = new_value;
			break;
		default:
			break;
	}
} /* ---------------------------------------------------------------------- */

/*
 * Load options from a file.
 */
static void load_options_from_file(const char * filename) {
	struct option_struct * current_option;
	FILE * file;
	char line[OPTIONS_LINE_SIZE];
	char * line_begin;

	file = fopen(filename, "r");
	if (file == NULL) {
		fprintf(stderr, _("Error opening file \"%s\" for reading: %s"), filename, strerror(errno));
		return;
	}

	while (!feof(file)) {
		memset(line, 0, sizeof(line));
		if (fgets(line, sizeof(line), file) == NULL) {
			if (!feof(file)) {
				fprintf(stderr, _("Error reading from file \"%s\": %s"), filename, strerror(errno));
			}
			break;
		}

		line_begin = line;

		/* Trim whitespace from line */
		while ((strlen(line_begin) > 0) && isspace(line_begin[strlen(line_begin)-1])) {
			line_begin[strlen(line_begin)-1] = '\0';
		}
		while ((isspace(*line_begin)) && (strlen(line_begin) > 0)) {
			line_begin++;
		}

		if (*line_begin == '#') {
			/* Skip comment lines */
			continue;
		}

		/* Look for option this line changes */
		current_option = options;
		while (current_option->option != Q_OPTION_NULL) {
			if (strstr(line, current_option->name) == line) {
				if ((line[strlen(current_option->name)] == '=') ||
					(isspace(line[strlen(current_option->name)]))) {

					/* Valid option */
					set_option(current_option, line + strlen(current_option->name));
				}
			}
			check_option(current_option);
			current_option++;
		}
	}

	fclose(file);
} /* ---------------------------------------------------------------------- */

static char home_directory_options_filename[FILENAME_SIZE];

/*
 * Get the full filename to "$HOME/.qodem/qodemrc"
 */
char * get_options_filename() {
	return home_directory_options_filename;
} /* ---------------------------------------------------------------------- */

/*
 * Find the option_struct from the Q_OPTION value
 */
static struct option_struct * find_option(const Q_OPTION option) {
	struct option_struct * current_option;
	
	/* Look for option this line changes */
	current_option = options;
	while (current_option->option != Q_OPTION_NULL) {
		if (current_option->option == option) {
			return current_option;
		}
		current_option++;
	}
	return NULL;
} /* ---------------------------------------------------------------------- */


/*
 * Load options from all the files.  We search the following:
 *     $HOME/.qodemrc
 *     INSTALL_DIR/qodemrc
 *     /etc/qodemrc
 *     /usr/lib/qodem/qodemrc
 *     /usr/local/lib/qodem/qodemrc
 */
void load_options() {
	int i;
	int state;
	char * begin;
	unsigned char ch = 0;
	char * filenames[] = {
		INSTALL_DIR "/qodemrc",
		"/etc/qodemrc",
		"/usr/lib/qodem/qodemrc",
		"/usr/local/lib/qodem/qodemrc",

		/* List $HOME last so it overrides everything else */
		"$HOME/.qodem/qodemrc",
		NULL
	};
	char * current_filename;
	char * env_string;
	char * substituted_filename;
	char * working_dir;
	struct option_struct * current_option;
	char lang_value[32];
	char * lang_default;

	/* Set default values. */
	current_option = options;
	while (current_option->option != Q_OPTION_NULL) {
		if (current_option->value != NULL) {
			Xfree(current_option->value, __FILE__, __LINE__);
		}
		current_option->value = Xstrdup(current_option->default_value, __FILE__, __LINE__);

		/* Translate option help text to local language */
		current_option->comment = _(current_option->comment);
		current_option++;
	}

	env_string = getenv("HOME");
	if (env_string == NULL) {
		env_string = "";
	} else {
		/* Check for .qodem directory */
		working_dir = substitute_string("$HOME/.qodem", "$HOME", env_string);
		if (access(working_dir, F_OK) != 0) {
			printf(_(""
"Qodem needs to create a directory to store its internal data files.\n"
"The default directory is %s.  Should I create this directory now? [Y/n] "), working_dir);
			i = getchar();
			if ((tolower(i) == 'y') || (i == '\r') || (i == '\n')) {
				i = mkdir(working_dir, S_IRUSR | S_IWUSR | S_IXUSR);
				if (i < 0) {
					printf(_(""
"Could not create the directory %s.  You may have to specify full paths\n"
"when you load key bindings, phone books, etc.\n"), working_dir);
				} else {
					printf(_("Created directory %s.\n"), working_dir);

					/* Create the key bindings files */
					create_keybindings_files();

					/* Create the translation tables file */
					create_translate_table_file();

					/* Create the modem config file */
					create_modem_config_file();
				}
			} else {
				printf(_(""
"Will NOT create the directory %s.  You may have to specify full paths\n"
"when you load key bindings, phone books, etc.\n"), working_dir);
			}

			/* Clear any other characters waiting in stdin */
			purge_stdin();

			printf(_("Press any key to continue...\n"));
			getchar();
		}

		/* Free leak */
		Xfree(working_dir, __FILE__, __LINE__);
	}

	/* Special check: $HOME/.qodem/qodemrc */
	/* Sustitute for $HOME */
	substituted_filename = substitute_string("$HOME/.qodem/qodemrc", "$HOME", env_string);
	if (access(substituted_filename, F_OK) != 0) {

		/* UTF-8 locale */

		lang_default = getenv("LANG");
		if (lang_default == NULL) {
			lang_default = get_option(Q_OPTION_UTF8_LANG);
		} else {
			if (strstr(lang_default, "UTF-8") == NULL) {
				lang_default = get_option(Q_OPTION_UTF8_LANG);
			}
		}

		printf(_(""
"Qodem needs to set the LANG environment variable when it connects to remote systems.\n"
"The current value for UTF-8 systems is '%s'.  Press Enter to use this value, or type\n"
"in a new value:  "), lang_default);

		fgets(lang_value, sizeof(lang_value) - 1, stdin);
		if (strlen(lang_value) > 0) {
			while (isspace(lang_value[strlen(lang_value) - 1])) {
				lang_value[strlen(lang_value) - 1] = 0;
			}
		}
		begin = lang_value;
		while ((strlen(begin) > 0) && (isspace(*begin))) {
			begin++;
		}
		if (strlen(begin) > 0) {
			set_option(find_option(Q_OPTION_UTF8_LANG), begin);
		} else {
			set_option(find_option(Q_OPTION_UTF8_LANG), lang_default);
		}

		/* 8-bit code page locale */

		lang_default = get_option(Q_OPTION_ISO8859_LANG);

		printf(_(""
"Qodem needs to set the LANG environment variable when it connects to remote systems.\n"
"The current value for non-UTF-8 systems is '%s'.  Press Enter to use this value, or type\n"
"in a new value:  "), lang_default);

		fgets(lang_value, sizeof(lang_value) - 1, stdin);
		if (strlen(lang_value) > 0) {
			while (isspace(lang_value[strlen(lang_value) - 1])) {
				lang_value[strlen(lang_value) - 1] = 0;
			}
		}
		begin = lang_value;
		while ((strlen(begin) > 0) && (isspace(*begin))) {
			begin++;
		}
		if (strlen(begin) > 0) {
			set_option(find_option(Q_OPTION_ISO8859_LANG), begin);
		} else {
			set_option(find_option(Q_OPTION_ISO8859_LANG), lang_default);
		}

		/* Save options */
		if (save_options(substituted_filename) == Q_FALSE) {
			printf(_("Error saving default options to %s\n"), substituted_filename);
			printf(_("Press any key to continue...\n"));
			getchar();
		}
	}
	/* Free leak */
	Xfree(substituted_filename, __FILE__, __LINE__);

	/* Special check: $HOME/.qodem/scripts */
	/* Sustitute for $HOME */
	substituted_filename = substitute_string("$HOME/.qodem/scripts", "$HOME", env_string);
	if (access(substituted_filename, F_OK) != 0) {
		/* Try to create it */
		mkdir(substituted_filename, S_IRUSR | S_IWUSR | S_IXUSR);
	}
	/* Free leak */
	Xfree(substituted_filename, __FILE__, __LINE__);

	/* Special check: $HOME/.qodem/scripts/script.stderr */
	/* Sustitute for $HOME */
	substituted_filename = substitute_string("$HOME/.qodem/scripts/script.stderr", "$HOME", env_string);
	if (access(substituted_filename, F_OK) != 0) {
		/* Try to create it */
		mkfifo(substituted_filename, S_IRUSR | S_IWUSR);
	}
	/* Free leak */
	Xfree(substituted_filename, __FILE__, __LINE__);
	
	i = 0;
	for (current_filename = filenames[i]; current_filename != NULL; i++, current_filename = filenames[i]) {
		/* Sustitute for $HOME */
		substituted_filename = substitute_string(current_filename, "$HOME", env_string);

		/* Save that last filename so we can use it for Alt-N Configuration later */
		if (filenames[i + 1] == NULL) {
			sprintf(home_directory_options_filename, "%s", substituted_filename);
		}

		/* Check existence of file */
		if (access(substituted_filename, F_OK | R_OK) == 0) {
			load_options_from_file(substituted_filename);
		}
		/* Free leak */
		Xfree(substituted_filename, __FILE__, __LINE__);

	}

	/* Check for working directory */
	working_dir = substitute_string(get_option(Q_OPTION_WORKING_DIR), "$HOME", env_string);
	if (access(working_dir, F_OK) != 0) {
		printf(_(""
"Qodem needs to create a directory to store user files such\n"
"as downloaded and uploaded files, session logs, capture files, etc.\n"
"The default directory is %s.  Should I create this directory now? [Y/n] "), working_dir);
		i = getchar();
		if ((tolower(i) == 'y') || (i == '\r') || (i == '\n')) {
			i = mkdir(working_dir, S_IRUSR | S_IWUSR | S_IXUSR);
			if (i < 0) {
				printf(_(""
"Could not create the directory %s.  You may have to specify full paths\n"
"when you download files, enable capture/log, etc.\n"), working_dir);
			} else {
				printf(_("Created directory %s.\n"), working_dir);
			}
		} else {
			printf(_(""
"Will NOT create the directory %s.  You may have to specify full paths\n"
"when you download files, enable capture/log, etc.\n"), working_dir);
		}

		/* Clear any other characters waiting in stdin */
		purge_stdin();

		printf(_("Press any key to continue...\n"));
		getchar();
	}
	/* Free leak */
	Xfree(working_dir, __FILE__, __LINE__);
	
	/*
	 * Special-case options.  For each one, reset to default and then re-load from file.
	 */

	q_status.idle_timeout		= 0;
	if (get_option(Q_OPTION_IDLE_TIMEOUT) != NULL) {
		q_status.idle_timeout = atoi(get_option(Q_OPTION_IDLE_TIMEOUT));
	}

	q_screensaver_timeout		= 0;
	if (get_option(Q_OPTION_SCREENSAVER_TIMEOUT) != NULL) {
		q_screensaver_timeout = atoi(get_option(Q_OPTION_SCREENSAVER_TIMEOUT));
	}

	q_keepalive_timeout		= 0;
	if (get_option(Q_OPTION_KEEPALIVE_TIMEOUT) != NULL) {
		q_keepalive_timeout = atoi(get_option(Q_OPTION_KEEPALIVE_TIMEOUT));
	}
	if (strlen(get_option(Q_OPTION_KEEPALIVE_BYTES)) > 0) {
		memset(q_keepalive_bytes, 0, sizeof(q_keepalive_bytes));
		state = 0;
		begin = get_option(Q_OPTION_KEEPALIVE_BYTES);
		for (i = 0; i < strlen(get_option(Q_OPTION_KEEPALIVE_BYTES)); i++) {
			if (q_keepalive_bytes_n == sizeof(q_keepalive_bytes)) {
				/* Max length */
				break;
			}
			switch (state) {
			case 0:
				if (begin[i] == '\\') {
					/* Embedded hex chracter */
					ch = 0;
					state = 1;
					continue;
				}
				q_keepalive_bytes[q_keepalive_bytes_n] = begin[i];
				q_keepalive_bytes_n++;
				break;
			case 1:
				if (tolower(begin[i]) == 'x') {
					/* Embedded hex chracter */
					ch = 0;
					state = 2;
					continue;
				}
				q_keepalive_bytes[q_keepalive_bytes_n] = '\\';
				q_keepalive_bytes_n++;
				state = 0;
				break;
			case 2:
				ch *= 16;
				if ((tolower(begin[i]) >= 'a') && (tolower(begin[i]) <= 'f')) {
					ch += begin[i] - 'a' + 16;
				} else if ((begin[i] >= '0') && (begin[i] <= '9')) {
					ch += begin[i] - '0';
				} else {
					q_keepalive_bytes[q_keepalive_bytes_n] = '\\';
					q_keepalive_bytes_n++;
					if (q_keepalive_bytes_n < sizeof(q_keepalive_bytes)) {
						q_keepalive_bytes[q_keepalive_bytes_n] = begin[i - 1];
						q_keepalive_bytes_n++;
					}
					state = 0;
					i--;
					continue;
				}
				state = 3;
				continue;

			case 3:
				ch *= 16;
				if ((tolower(begin[i]) >= 'a') && (tolower(begin[i]) <= 'f')) {
					ch += begin[i] - 'a' + 16;
				} else if ((begin[i] >= '0') && (begin[i] <= '9')) {
					ch += begin[i] - '0';
				} else {
					i--;
				}
				q_keepalive_bytes[q_keepalive_bytes_n] = ch;
				q_keepalive_bytes_n++;
				state = 0;
				continue;
			}
		}
	}

	q_status.capture_type		= Q_CAPTURE_TYPE_NORMAL;
	if (strcasecmp(get_option(Q_OPTION_CAPTURE_TYPE), "raw") == 0) {
		q_status.capture_type = Q_CAPTURE_TYPE_RAW;
	}
	if (strcasecmp(get_option(Q_OPTION_CAPTURE_TYPE), "html") == 0) {
		q_status.capture_type = Q_CAPTURE_TYPE_HTML;
	}

	q_status.sound = Q_FALSE;
	q_status.beeps = Q_FALSE;
	q_status.ansi_music = Q_FALSE;
	if (strcasecmp(get_option(Q_OPTION_SOUNDS_ENABLED), "true") == 0) {
		q_status.sound = Q_TRUE;
		q_status.beeps = Q_TRUE;

		if (strcasecmp(get_option(Q_OPTION_ANSI_MUSIC), "true") == 0) {
			q_status.ansi_music = Q_TRUE;
		}
	}

	q_status.zmodem_autostart	= Q_TRUE;
	if (strcasecmp(get_option(Q_OPTION_ZMODEM_AUTOSTART), "false") == 0) {
		q_status.zmodem_autostart = Q_FALSE;
	}
	q_status.zmodem_zchallenge	= Q_FALSE;
	if (strcasecmp(get_option(Q_OPTION_ZMODEM_ZCHALLENGE), "true") == 0) {
		q_status.zmodem_zchallenge = Q_TRUE;
	}
	q_status.zmodem_escape_ctrl	= Q_FALSE;
	if (strcasecmp(get_option(Q_OPTION_ZMODEM_ESCAPE_CTRL), "true") == 0) {
		q_status.zmodem_escape_ctrl = Q_TRUE;
	}

	q_status.kermit_autostart	= Q_TRUE;
	if (strcasecmp(get_option(Q_OPTION_KERMIT_AUTOSTART), "false") == 0) {
		q_status.kermit_autostart = Q_FALSE;
	}
	q_status.kermit_robust_filename	= Q_FALSE;
	if (strcasecmp(get_option(Q_OPTION_KERMIT_ROBUST_FILENAME), "true") == 0) {
		q_status.kermit_robust_filename = Q_TRUE;
	}
	q_status.kermit_streaming	= Q_TRUE;
	if (strcasecmp(get_option(Q_OPTION_KERMIT_STREAMING), "false") == 0) {
		q_status.kermit_streaming = Q_FALSE;
	}
	q_status.kermit_uploads_force_binary	= Q_TRUE;
	if (strcasecmp(get_option(Q_OPTION_KERMIT_UPLOADS_FORCE_BINARY), "false") == 0) {
		q_status.kermit_uploads_force_binary = Q_FALSE;
	}
	q_status.kermit_downloads_convert_text	= Q_FALSE;
	if (strcasecmp(get_option(Q_OPTION_KERMIT_DOWNLOADS_CONVERT_TEXT), "true") == 0) {
		q_status.kermit_downloads_convert_text = Q_TRUE;
	}

	q_status.assume_80_columns	= Q_TRUE;
	if (strcasecmp(get_option(Q_OPTION_80_COLUMNS), "false") == 0) {
		q_status.assume_80_columns = Q_FALSE;
	}

	q_status.ansi_animate		= Q_FALSE;
	if (strcasecmp(get_option(Q_OPTION_ANSI_ANIMATE), "true") == 0) {
		q_status.ansi_animate = Q_TRUE;
	}

} /* ---------------------------------------------------------------------- */
