/* $Id: scrollback.c,v 1.86 2010/09/10 22:55:50 lamonte Exp $
 *
 * scrollback.c
 *
 * Additional acknowledgements, news, and updates to the Qodem Project
 * are available at http://qodem.sourceforge.net .
 *
 * This module is licensed under the GNU General Public License Version 2.
 * Please see the file "COPYING" in this directory for more information
 * about the GNU General Public License Version 2.
 *
 *     Copyright (C) 2003  Kevin Lamonte
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	 USA
 *
 *
 */

/*
 * This module contains the scrollback buffer handling code.
 *
 * The scrollback buffer is a linked list of struct q_scrolline_struct *'s
 * starting at q_scrollback_buffer.  However, the behavior of that list
 * is a little unusual:
 *
 *     1)  The visible portion of the scollback buffer is q_scrollback_position
 *	   and goes BACKWARD to find_top_scrollback_line().
 *
 *     2)  The current cursor position (q_status.cursor_x/y) is on
 *	   q_scrollback_current.
 *
 *     3)  The scrollback buffer is wider and taller than the visible screen.
 *
 * Visibly, this looks like:
 *
 *				   | <-------  q_scrolline_struct.length  ------> |
 *
 *				   | <----------  WIDTH	 ----------> |
 *
 * q_scrollback_buffer	      ---> .---------------------------------.--------... |
 *				   |---------------------------------|--------... |
 *				   |---------------------------------|--------... |
 * find_top_scrollback_line() ---> |0********************************|--------... |
 *				   |*********************************|--------... |
 *				   |*********************************|--------... |
 *				   |*********************************|--------... |
 * q_scrollback_current	      ---> |******X**************************|--------... |
 *				   |*********************************|--------... |
 *				   |*********************************|--------... |
 * q_scrollback_position      ---> |*********************************|--------... |
 *			 /-------> |%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|
 *	      STATUS_HEIGHT	   |%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|
 *			 \-------> |%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|
 *				   |---------------------------------|--------... |
 * q_scrollback_last	      ---> ^---------------------------------^--------... ^
 *				    ^	  ^
 *				    |	  |
 *				    |	  ^--- "X" is on row=cursor_y, col=cursor_x
 *				    |
 *				    ^--- "0" is on row=0, col=0
 *
 *
 * Why this design?  Mainly because if the window is re-sized in
 * Q_STATE_CONSOLE state I want to make sure the last line is still
 * visible directly above the status line(s).
 *
 * Note that the only time that q_scrollback_last cannot be equal to
 * q_scrollback_position is when state is Q_STATE_SCROLLBACK.
 *
 * I almost want to add a variable origin (0,0) so that the screen
 * could pan left to right.  But I don't know any of any terminals
 * that require that.  Still, it would bring new life to "scroll
 * lock".  :)
 */

#include "common.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>			/* strncpy() */
#include <assert.h>
#include <errno.h>
#include "qodem.h"
#include "screen.h"
#include "forms.h"
#include "options.h"
#include "states.h"
#include "status.h"
#include "codepage.h"
#include "script.h"
#include "scrollback.h"

/* The scrollback buffer. */
struct q_scrolline_struct * q_scrollback_buffer = NULL;

/* The last line of the scrollback buffer. */
struct q_scrolline_struct * q_scrollback_last = NULL;

/* The current edit line of the scrollback buffer. */
struct q_scrolline_struct * q_scrollback_current = NULL;

/* The current view position in the scrollback buffer.
   This position is the BOTTOM on the screen, so when we render we
   count from here and go UP until we run out of available rows
   to render to. */
struct q_scrolline_struct * q_scrollback_position = NULL;

/* The maximum size of the scrollback buffer. Default 20000. */
static int scrollback_max = 20000;

/*
 * Find the scrollback line that corresponds to the top line
 * of the screen.
 */
static struct q_scrolline_struct * find_top_scrollback_line() {
	struct q_scrolline_struct * line;
	int row;

	/* Start at the bottom */
	row = HEIGHT - 1;

	/* Skip the status line */
	if (q_status.status_visible == Q_TRUE) {
		row -= STATUS_HEIGHT;
	}

	/* Let's assert that row > 0.  Konsole and xterm won't
	   let the window size reach zero so this should be a
	   non-issue. */
	assert(row > 0);

	/* Count the lines available */
	line = q_scrollback_position;
	while (row >= 0) {
		if (line->prev == NULL) {
			break;
		}
		line = line->prev;
		row--;
	}

	if ((row<0) && (line->next != NULL)) {
		line = line->next;
	}

	return line;
} /* ---------------------------------------------------------------------- */

/*
 * insert_scrollback_line - initialize a new line for the scrollback
 * buffer.  The line is inserted before insertion_point.
 */
static void insert_scrollback_line(struct q_scrolline_struct * insertion_point) {
	struct q_scrolline_struct * new_line;
	int i;

	assert (insertion_point != NULL);

	new_line = (struct q_scrolline_struct *)Xmalloc(sizeof(struct q_scrolline_struct), __FILE__, __LINE__);
	memset(new_line, 0, sizeof(struct q_scrolline_struct));
	for (i = 0; i < Q_MAX_LINE_LENGTH; i++) {
		new_line->chars[i] = ' ';
	}
	new_line->dirty = Q_TRUE;
	/* All new lines are always single-width / single-height */
	new_line->double_width = Q_FALSE;
	new_line->double_height = 0;

	if (q_status.reverse_video == Q_TRUE) {
		for (i = 0; i < Q_MAX_LINE_LENGTH; i++) {
			new_line->colors[i] = Q_A_REVERSE | scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
		}
		/* Increase width and pad with spaces so that inverse video
		   appears everwhere. */
		new_line->length = WIDTH;
	}

	if (q_scrollback_buffer == NULL) {
		q_scrollback_buffer = new_line;
		q_scrollback_position = new_line;
		q_scrollback_last = new_line;
		q_scrollback_current = new_line;
	} else {
		new_line->prev = insertion_point->prev;
		new_line->next = insertion_point;
		insertion_point->prev = new_line;
		/*
		 * ASCII downloads and the console itself both update the
		 * scrollback and need to render the new line.
		 */
		if (	(q_program_state == Q_STATE_CONSOLE) ||
			(q_program_state == Q_STATE_SCRIPT_EXECUTE) ||
			(q_program_state == Q_STATE_DOWNLOAD)
		) {
			q_scrollback_position = q_scrollback_position->prev;
		}
	}

	if (
		(q_status.scrollback_lines == scrollback_max) ||
		((q_status.scrollback_enabled == Q_FALSE) && (q_status.scrollback_lines > HEIGHT - STATUS_HEIGHT - 1))
	) {

		/* Roll the bottom line off the buffer. */
		new_line = q_scrollback_last;
		q_scrollback_last = new_line->prev;
		q_scrollback_last->next = NULL;
		Xfree(new_line, __FILE__, __LINE__);

		if (q_scrollback_position == new_line) {
			q_scrollback_position = q_scrollback_position->prev;
		}
	} else {
		q_status.scrollback_lines++;
	}
} /* ---------------------------------------------------------------------- */

/*
 * new_scrollback_line - initialize a new line for the scrollback
 * buffer.  The line is appended to the buffer, becoming the new
 * q_scrollback_last.
 */
void new_scrollback_line() {
	struct q_scrolline_struct * new_line = NULL;
	struct q_scrolline_struct * top_line = NULL;
	int i;

	new_line = (struct q_scrolline_struct *)Xmalloc(sizeof(struct q_scrolline_struct), __FILE__, __LINE__);
	memset(new_line, 0, sizeof(struct q_scrolline_struct));
	for (i = 0; i < Q_MAX_LINE_LENGTH; i++) {
		new_line->chars[i] = ' ';
	}
	new_line->dirty = Q_TRUE;
	/* All new lines are always single-width / single-height */
	new_line->double_width = Q_FALSE;
	new_line->double_height = 0;

	if (q_status.reverse_video == Q_TRUE) {
		for (i = 0; i < Q_MAX_LINE_LENGTH; i++) {
			new_line->colors[i] = Q_A_REVERSE | scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
		}
		/* Increase width and pad with spaces so that inverse video
		   appears everwhere. */
		new_line->length = WIDTH;
	}

	if (q_status.emulation == Q_EMUL_DEBUG) {
		/* DEBUG emulation plays tricks with the scrollback buffer.
		 * If I don't explicitly set the color the cursor will
		 * disappear.
		 */
		for (i = 0; i < Q_MAX_LINE_LENGTH; i++) {
			new_line->colors[i] = Q_A_REVERSE | scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
		}
	}

	if (q_scrollback_buffer == NULL) {
		q_scrollback_buffer = new_line;
		q_scrollback_position = new_line;
		q_scrollback_last = new_line;
		q_scrollback_current = new_line;
	} else {
		top_line = find_top_scrollback_line();

		new_line->prev = q_scrollback_last;
		q_scrollback_last->next = new_line;
		q_scrollback_last = new_line;
		/*
		 * ASCII downloads and the console itself both update the
		 * scrollback and need to render the new line.
		 *
		 * Also, debug_reset() needs to render its new lines.
		 */
		if ((q_program_state == Q_STATE_CONSOLE) ||
			(q_program_state == Q_STATE_DOWNLOAD) ||
			(q_program_state == Q_STATE_SCRIPT_EXECUTE) ||
			(q_program_state == Q_STATE_EMULATION_MENU)
		) {
			q_scrollback_position = new_line;
		}
	}

	if (
		(q_status.scrollback_lines == scrollback_max) ||
		((q_status.scrollback_enabled == Q_FALSE) && (q_status.scrollback_lines > HEIGHT - STATUS_HEIGHT - 1))
	) {

		if (q_status.scrollback_enabled == Q_TRUE) {
			/* Roll the top line off the buffer. */
			new_line = q_scrollback_buffer;
			q_scrollback_buffer = new_line->next;
			q_scrollback_buffer->prev = NULL;
			Xfree(new_line, __FILE__, __LINE__);
		} else {
			/* Roll the top line in the visible area off the buffer. */
			top_line->next->prev = top_line->prev;
			if (top_line->prev != NULL) {
				top_line->prev->next = top_line->next;
			}
			Xfree(top_line, __FILE__, __LINE__);

		}

	} else {
		q_status.scrollback_lines++;
	}
} /* ---------------------------------------------------------------------- */

/* Special case for VT100 line wrapping. */
static Q_BOOL vt100_wrap_line_flag = Q_FALSE;

/*
 * wrap_current_line -
 *
 * The code to wrap a line.  It has two different places it might be called,
 * so now I've got to separate it into a function.
 */
static void wrap_current_line() {
	/* Wrap the line */
	if (q_scrollback_current->next == NULL) {
		new_scrollback_line();
	}
	q_scrollback_current = q_scrollback_current->next;
	if (q_status.cursor_y < HEIGHT - STATUS_HEIGHT - 1) {
		q_status.cursor_y++;
	}
	q_status.cursor_x = 0;

	/* Capture */
	if (q_status.capture == Q_TRUE) {
		if (q_status.capture_type == Q_CAPTURE_TYPE_HTML) {
			/* HTML */
			fprintf(q_status.capture_file, "\n");
		} else if (q_status.capture_type == Q_CAPTURE_TYPE_NORMAL) {
			/* Normal */
			fprintf(q_status.capture_file, "\n");
		}
		fflush(q_status.capture_file);
		q_status.capture_flush_time = time(NULL);
		q_status.capture_x = 0;
	}
} /* ---------------------------------------------------------------------- */

/*
 * print_character
 */
void print_character(const wchar_t character) {
	static attr_t old_color = 0xdeadbeef;
	Q_BOOL color_changed = Q_FALSE;
	int right_margin = WIDTH - 1;
	Q_BOOL wrap_the_line = Q_FALSE;
	int i;

	if (q_scrollback_current->length < q_status.cursor_x) {
		for (i = q_scrollback_current->length; i < q_status.cursor_x; i++) {
			q_scrollback_current->chars[i] = ' ';
			q_scrollback_current->colors[i] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
			q_scrollback_current->length = q_status.cursor_x;
		}
	}

	/*
	 * I want a const character in the API, but it's convenient
	 * for flow control to change character.
	 */
	wchar_t character2 = character;

	/* Initialize old_color */
	if (old_color == 0xdeadbeef) {
		old_color = q_current_color;
		color_changed = Q_FALSE;
	}

	/* BEL */
	if (character2 == 0x07) {
		screen_beep();
		return;
	}

	/* NUL */
	if (character2 == 0x00) {
		if (q_status.display_null == Q_TRUE) {
			character2 = ' ';
		} else {
			return;
		}
	}

	/* A character will be printed, mark the line dirty */
	q_scrollback_current->dirty = Q_TRUE;

	/* Pass the character to a script if we're running one */
	if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
		script_print_character(character2);
	}
	
	/* This isn't the prettiest logic, but whatever */
	switch (q_status.emulation) {
	case Q_EMUL_ANSI:
	case Q_EMUL_AVATAR:
	case Q_EMUL_TTY:
		/* BBS-ish emulations:  check the assume_80_columns flag */
		if (q_status.assume_80_columns == Q_TRUE) {
			right_margin = 79;
		}
		break;
	default:
		/* VT100-ish emulations:  check the actual right margin value */
		if (q_emulation_right_margin > 0) {
			right_margin = q_emulation_right_margin;
		}
		break;
	}
	if (q_scrollback_current->double_width == Q_TRUE) {
		right_margin = ((right_margin + 1) / 2) - 1;
	}

	/* Check the unusually-complicated line wrapping conditions... */
	if (q_status.cursor_x == right_margin) {

		/* This case happens when:  the cursor was already
		   on the right margin (either through printing or by
		   an explicit placement command), and a character
		   was printed. */

		if ((q_status.emulation == Q_EMUL_VT100) ||
		    (q_status.emulation == Q_EMUL_VT102) ||
		    (q_status.emulation == Q_EMUL_VT220) ||
		    (q_status.emulation == Q_EMUL_LINUX) ||
		    (q_status.emulation == Q_EMUL_LINUX_UTF8) ||
		    (q_status.emulation == Q_EMUL_XTERM) ||
		    (q_status.emulation == Q_EMUL_XTERM_UTF8)
		) {
			/* Special case for VT100: the line wraps
			   only when a new character arrives AND the cursor
			   is already on the right margin AND has placed
			   a character in its cell.  Easier to see than
			   to explain. */

			if (q_status.line_wrap == Q_TRUE) {

				if (vt100_wrap_line_flag == Q_FALSE) {

					/* This block marks the case that we are
					   in the margin and the first character
					   has been received and printed. */
					vt100_wrap_line_flag = Q_TRUE;

				} else {
					/* This block marks the case that we are
					   in the margin and the second character
					   has been received and printed. */
					vt100_wrap_line_flag = Q_FALSE;
					wrap_the_line = Q_FALSE;

					wrap_current_line();
				}
			} else {
				/*
				 * Line wrap is OFF so do not wrap the line.
				 */
				wrap_the_line = Q_FALSE;
			}

		} else {
			/* We are NOT in VT100 emulation... */

			if (q_status.line_wrap == Q_TRUE) {
				/* Regular NON-VT100 case.  The cursor moves
				   to the first column of the next line as soon
				   as the character has been printed in the
				   right margin column. */

				wrap_the_line = Q_TRUE;
			} else {
				/*
				 * Line wrap is OFF so do not wrap the line.
				 *
				 * Also, Keep the cursor where it is on the
				 * right margin.
				 */
				wrap_the_line = Q_FALSE;

			}

		} /* if (q_status.emulation == Q_EMUL_VT100) */

	} else if (q_status.cursor_x <= right_margin) {
		/* This is the normal case: a character came in and
		   was printed to the left of the right margin column. */
		wrap_the_line = Q_FALSE;

		/* Turn off VT100 special-case flag */
		vt100_wrap_line_flag = Q_FALSE;
	}

	/* Print the character */
	if (q_status.cursor_x < q_scrollback_current->length) {

		/* Insert mode special case */
		if (q_status.insert_mode == Q_TRUE) {
			memmove(&q_scrollback_current->chars[q_status.cursor_x + 1],
				&q_scrollback_current->chars[q_status.cursor_x],
				sizeof(wchar_t) * (Q_MAX_LINE_LENGTH - q_status.cursor_x - 1));

			memmove(&q_scrollback_current->colors[q_status.cursor_x + 1],
				&q_scrollback_current->colors[q_status.cursor_x],
				sizeof(attr_t) * (Q_MAX_LINE_LENGTH - q_status.cursor_x - 1));

			q_scrollback_current->chars[q_status.cursor_x] = character2;
			q_scrollback_current->colors[q_status.cursor_x] = q_current_color;
			if (q_scrollback_current->length < Q_MAX_LINE_LENGTH) {
				q_scrollback_current->length++;
			}
		} else {
			/* Replace an existing character */
			q_scrollback_current->colors[q_status.cursor_x] = q_current_color;
			q_scrollback_current->chars[q_status.cursor_x] = character2;
		}
	} else {
		/* New character on the line */
		q_scrollback_current->colors[q_scrollback_current->length] = q_current_color;
		q_scrollback_current->chars[q_scrollback_current->length] = character2;
		q_scrollback_current->length++;
	}

	/* Check the color */
	if (q_current_color != old_color) {
		color_changed = Q_TRUE;
		old_color = q_current_color;
	}

	/* Capture */
	if (q_status.capture == Q_TRUE) {
		/* Insert spaces to line up capture */
		if (q_status.cursor_x > q_status.capture_x) {
			for (i = 0; i < q_status.cursor_x - q_status.capture_x; i++) {
				if (q_status.capture_type == Q_CAPTURE_TYPE_HTML) {
					fprintf(q_status.capture_file, "&nbsp;");
				} else {
					fprintf(q_status.capture_file, " ");
				}
				if (q_scrollback_current->double_width == Q_TRUE) {
					fprintf(q_status.capture_file, " ");
				}
			}
			q_status.capture_x = q_status.cursor_x;
		}

		if (q_status.capture_type == Q_CAPTURE_TYPE_HTML) {
			/* HTML */
			if (color_changed == Q_TRUE) {
				fprintf(q_status.capture_file, "</font><font %s>", color_to_html(q_current_color));
			}
			if (character2 == ' ') {
				fprintf(q_status.capture_file, "&nbsp;");
			} else if (character2 == '<') {
				fprintf(q_status.capture_file, "&lt;");
			} else if (character2 == '>') {
				fprintf(q_status.capture_file, "&gt;");
			} else if (character2 < 0x7F) {
				fprintf(q_status.capture_file, "%c", character2);
			} else {
				fprintf(q_status.capture_file, "&#%d;", character2);
			}
		} else if (q_status.capture_type == Q_CAPTURE_TYPE_NORMAL) {
			/* Normal */
			fprintf(q_status.capture_file, "%lc", character2);
		}
		q_status.capture_x++;

		/* Double-width */
		if (q_scrollback_current->double_width == Q_TRUE) {
			fprintf(q_status.capture_file, " ");
			q_status.capture_x++;
		}

		/* Flush if we haven't in a while */
		if (q_status.capture_flush_time < time(NULL)) {
			fflush(q_status.capture_file);
			q_status.capture_flush_time = time(NULL);
		}

	}

	/* Increment horizontal */
	if (vt100_wrap_line_flag == Q_FALSE) {
		q_status.cursor_x++;

		/* Use the right margin instead of WIDTH */
		/* if (q_status.cursor_x == WIDTH) { */
		if (q_status.cursor_x > right_margin) {
			q_status.cursor_x--;
		}
	}

	/* Wrap if necessary */
	if (wrap_the_line == Q_TRUE) {
		/* NON-VT100 "normal" case:  add a character and then wrap the line. */
		wrap_current_line();
	} /* if (wrap_the_line == Q_TRUE) */
} /* ---------------------------------------------------------------------- */

/*
 * clear_scrollback
 */
static void clear_scrollback() {
	struct q_scrolline_struct * line;
	struct q_scrolline_struct * line_next;
	struct q_scrolline_struct * top;

	top = find_top_scrollback_line();

	line = q_scrollback_buffer;
	while (line != top) {
		line_next = line->next;
		Xfree(line, __FILE__, __LINE__);
		q_status.scrollback_lines--;
		line = line_next;
	}

	/* Mark this the head. */
	top->prev = NULL;
	q_scrollback_buffer = top;

} /* ---------------------------------------------------------------------- */

/*
 * save_scrollback
 */
static Q_BOOL save_scrollback(const char * filename) {
	struct q_scrolline_struct * line;
	int i;
	FILE * file;
	time_t current_time;
	char time_string[TIME_STRING_LENGTH];
	char * new_filename = NULL;
	char notify_message[DIALOG_MESSAGE_SIZE];

	if (filename[0] != '/') {
		/* Relative path, prefix working directory */
		new_filename = (char *)Xmalloc(strlen(filename) +
			strlen(get_option(Q_OPTION_WORKING_DIR)) + 2, __FILE__, __LINE__);
		memset(new_filename, 0, strlen(filename) + strlen(get_option(Q_OPTION_WORKING_DIR)) + 2);

		strncpy(new_filename, get_option(Q_OPTION_WORKING_DIR), strlen(get_option(Q_OPTION_WORKING_DIR)));
		new_filename[strlen(new_filename)] = '/';
		strncpy(new_filename + strlen(new_filename), filename, strlen(filename));
		filename = new_filename;
	}

	file = fopen(filename, "a");
	if (file == NULL) {
		snprintf(notify_message, sizeof(notify_message), _("Error opening file \"%s\" for writing: %s"), new_filename, strerror(errno));
		notify_form(notify_message, 0);
		if (new_filename != NULL) {
			Xfree(new_filename, __FILE__, __LINE__);
		}
		return Q_FALSE;
	}
	time(&current_time);
	strftime(time_string, sizeof(time_string), _("Saved Scrollback Generated %a, %d %b %Y %H:%M:%S %z"), localtime(&current_time));
	fprintf(file, "* - * Qodem " Q_VERSION " " Q_VERSION_BRANCH " %s BEGIN * - *\n\n", time_string);

	/* Now loop from line onward */
	for (line=q_scrollback_buffer; line!=NULL; line=line->next) {
		for (i = 0; i < line->length; i++) {
			if (line->double_width == Q_TRUE) {
				if ((2 * i) >= WIDTH) {
					break;
				}
				fprintf(file, "%lc", line->chars[i]);
				fprintf(file, " ");
			} else {
				if (i >= WIDTH) {
					break;
				}
				fprintf(file, "%lc", line->chars[i]);
			}
		}
		fprintf(file, "\n");
	}

	fprintf(file, "\n* - * Qodem " Q_VERSION " " Q_VERSION_BRANCH " %s END * - *\n", time_string);
	fclose(file);
	if (new_filename != NULL) {
		Xfree(new_filename, __FILE__, __LINE__);
	}
	return Q_TRUE;
} /* ---------------------------------------------------------------------- */

/*
 * screen_dump
 */
Q_BOOL screen_dump(const char * filename) {
	struct q_scrolline_struct * line;
	int i;
	FILE * file;
	time_t current_time;
	char time_string[TIME_STRING_LENGTH];
	char * new_filename = NULL;
	char notify_message[DIALOG_MESSAGE_SIZE];

	file = open_workingdir_file(filename, &new_filename);
	if (file == NULL) {
		snprintf(notify_message, sizeof(notify_message), _("Error opening file \"%s\" for writing: %s"), new_filename, strerror(errno));
		notify_form(notify_message, 0);
		if (new_filename != NULL) {
			Xfree(new_filename, __FILE__, __LINE__);
		}
		return Q_FALSE;
	}
	time(&current_time);
	strftime(time_string, sizeof(time_string), _("Screen Dump Generated %a, %d %b %Y %H:%M:%S %z"), localtime(&current_time));
	fprintf(file, "* - * Qodem " Q_VERSION " %s BEGIN * - *\n\n", time_string);

	line = find_top_scrollback_line();

	/* Now loop from line onward */
	while (line != NULL) {
		for (i = 0; i < line->length; i++) {
			if (line->double_width == Q_TRUE) {
				if ((2 * i) >= WIDTH) {
					break;
				}
				fprintf(file, "%lc", line->chars[i]);
				fprintf(file, " ");
			} else {
				if (i >= WIDTH) {
					break;
				}
				fprintf(file, "%lc", line->chars[i]);
			}
		}
		fprintf(file, "\n");
		line = line->next;
	}

	fprintf(file, "\n* - * %s END * - *\n", time_string);
	fclose(file);
	if (new_filename != filename) {
		Xfree(new_filename, __FILE__, __LINE__);
	}
	return Q_TRUE;
} /* ---------------------------------------------------------------------- */

/*
 * scrollback_keyboard_handler
 */
void scrollback_keyboard_handler(const int keystroke, const int flags) {
	struct q_scrolline_struct * line;
	int i, j;
	int local_height;
	char * filename;
	char notify_message[DIALOG_MESSAGE_SIZE];

	local_height = HEIGHT - STATUS_HEIGHT;

	switch (tolower(keystroke)) {

	case 's':
		/* Save */
		q_cursor_on();
		filename = save_form(_("Scroll Save Filename"), _("saved_scrollback.txt"), Q_FALSE, Q_FALSE);
		q_cursor_off();
		if (filename != NULL) {
			qlog(_("Scrollback saved to file '%s'\n"), filename);
			if (save_scrollback(filename) == Q_FALSE) {
				snprintf(notify_message, sizeof(notify_message), _("Error saving to file \"%s\""), filename);
				notify_form(notify_message, 0);
			}
		}
		/* Return to console */
		q_scrollback_position = q_scrollback_last;
		switch_state(Q_STATE_CONSOLE);
		break;

	case 'c':
		/* Clear */
		clear_scrollback();
		/* Return to console */
		q_scrollback_position = q_scrollback_last;
		switch_state(Q_STATE_CONSOLE);
		break;

	default:
		break;

	}

	switch (keystroke) {

	case '`':
	case KEY_ESCAPE:
		/* Return to console */
		q_scrollback_position = q_scrollback_last;
		switch_state(Q_STATE_CONSOLE);
		break;

	case Q_KEY_UP:
		/* Make sure we need to scroll. This is pretty ugly but
		 it works so I won't bugger with it anymore. Same code
		 is in the KEY_PPAGE case too but j instead of i. */
		for (i = 0, line=q_scrollback_buffer; i < local_height; i++) {
			if (line == q_scrollback_position) {
				break;
			}
			if ((line != NULL) && (line->next != NULL)) {
				line = line->next;
			}
		}
		if (i != local_height) {
			break;
		}

		if ((q_scrollback_position != NULL) && (q_scrollback_position->prev != NULL)) {
			q_scrollback_position = q_scrollback_position->prev;
		}
		break;

	case Q_KEY_DOWN:
		if ((q_scrollback_position != NULL) && (q_scrollback_position->next != NULL)) {
			q_scrollback_position = q_scrollback_position->next;
		}
		break;

	case Q_KEY_END:
		q_scrollback_position = q_scrollback_last;
		break;

	case Q_KEY_PPAGE:
		for (i = 0; i < local_height; i++) {
			for (j=0, line=q_scrollback_buffer; j<local_height; j++) {
				if (line == q_scrollback_position) {
					break;
				}
				if ((line != NULL) && (line->next != NULL)) {
					line = line->next;
				}
			}
			if (j != local_height) {
				i = local_height;
				break;
			}
			if ((q_scrollback_position != NULL) && (q_scrollback_position->prev != NULL)) {
				q_scrollback_position = q_scrollback_position->prev;
			}
		}
		break;

	case Q_KEY_HOME:
		q_scrollback_position = q_scrollback_buffer;
		for (i = 1; i < local_height; i++) {
			if ((q_scrollback_position != NULL) && (q_scrollback_position->next != NULL)) {
				q_scrollback_position = q_scrollback_position->next;
			}
		}
		break;

	case Q_KEY_NPAGE:
		for (i = 0; i < local_height; i++) {
			if ((q_scrollback_position != NULL) && (q_scrollback_position->next != NULL)) {
				q_scrollback_position = q_scrollback_position->next;
			}
		}
		break;

	default:
		break;
	}

} /* ---------------------------------------------------------------------- */

/*
 * render_scrollback
 */
void render_scrollback(const int skip_lines) {
	struct q_scrolline_struct * line;
	int row;
	int renderable_lines;
	int i;

	row = HEIGHT - 1;

	/* Skip the status line */
	if (q_status.status_visible == Q_TRUE) {
		row -= STATUS_HEIGHT;
	}

	/* Let's assert that row > 0.  Konsole and xterm won't
	   let the window size reach zero so this should be a
	   non-issue. */
	assert(row > 0);

	/* If split_screen is enabled, we will skip some lines from the top. */
	row -= skip_lines;
	if (row < 0) {
		/* Sanity check.  I don't want to abort, so we'll just return. */
		fprintf(stderr, "render_scrollback() WARNING: Screen size is too small: height=%d width=%d status_height=%d skip_lines=%d\n", HEIGHT, WIDTH, STATUS_HEIGHT, skip_lines);
		return;
	}

	/* Count the lines available */
	renderable_lines = 0;
	line = q_scrollback_position;
	while (row >= 0) {
		renderable_lines++;
		if (line->prev == NULL) {
			break;
		}
		line = line->prev;
		row--;
	}

	/* At this point line should point to the line directly
	 * above find_top_scrollback_line() IF row is -1.
	 * If row is 0 or more then the scrollback buffer is not
	 * as large as the screen.
	 */
	if ((row<0) && (line->next != NULL)) {
		line = line->next;
	}

	/* Now loop from line onward */
	for (row = 0; row < renderable_lines; row++) {

		/*
		 * ncurses-based drawing needs to refresh the whole screen
		 * matter what.  ncurses itself handles its own dirty state.
		 * But I want to keep the stub for doing dirty-line handling
		 * here for the case of rendering to a larger scrollback
		 * region (ala Turbo Vision).
		 */
		if ((line->dirty == Q_TRUE) || Q_TRUE) {

			if ((line->length > 0) && (line->double_height < 2)) {
				for (i = 0; i < line->length; i++) {

					if (line->double_width == Q_TRUE) {
						if ((2 * i) >= WIDTH) {
							break;
						}
						screen_put_scrollback_char_yx(row, (2 * i), line->chars[i], line->colors[i]);
						screen_put_scrollback_char_yx(row, (2 * i) + 1, ' ', line->colors[i]);
					} else {
						if (i >= WIDTH) {
							break;
						}
						screen_put_scrollback_char_yx(row, i, line->chars[i], line->colors[i]);
					}

				}
			} else {
				screen_move_yx(row, 0);
			}

			/* Clear remainder of line */
			screen_clear_remaining_line();

			line->dirty = Q_FALSE;

		} /* if (line->dirty == Q_TRUE) */

		line = line->next;
	}

	for (row = renderable_lines; row < HEIGHT - STATUS_HEIGHT; row++) {
		screen_move_yx(row, 0);
		screen_clear_remaining_line();
	}
} /* ---------------------------------------------------------------------- */

/*
 * scrollback_refresh
 */
void scrollback_refresh() {
	char * scrollback_string;
	int status_left_stop;

	/* Put up the status line */
	screen_put_color_hline_yx(HEIGHT - 1, 0, cp437_chars[HATCH], WIDTH, Q_COLOR_STATUS);

	scrollback_string = _(" SCROLL-BACK                           S-Save  C-Clear  ESC/`-Exit  Lines:");
	status_left_stop = WIDTH - strlen(scrollback_string) - 6;
	if (status_left_stop <= 0) {
		status_left_stop = 0;
	} else {
		status_left_stop /= 2;
	}

	screen_put_color_printf_yx(HEIGHT - 1, status_left_stop, Q_COLOR_STATUS, "%s %'-5u ", scrollback_string, q_status.scrollback_lines);

	/* Add arrows on the status line */
	if (q_scrollback_position != q_scrollback_last) {
		/* Down arrow - more lines are below */
		screen_put_color_char_yx(HEIGHT - 1, status_left_stop + 28, cp437_chars[DOWNARROW], Q_COLOR_STATUS);
	}
	if (find_top_scrollback_line() != q_scrollback_buffer) {
		/* Up arrow - more lines are above */
		screen_put_color_char_yx(HEIGHT - 1, status_left_stop + 27, cp437_chars[UPARROW], Q_COLOR_STATUS);
	}

	/* Render scrollback */
	render_scrollback(0);

	if (q_scrollback_current->double_width == Q_TRUE) {
		screen_move_yx(q_status.cursor_y, (2 * q_status.cursor_x));
	} else {
		screen_move_yx(q_status.cursor_y, q_status.cursor_x);
	}
	screen_flush();
} /* ---------------------------------------------------------------------- */

/*
 * Scroll a region between region_top and region_bottom, inclusive, up
 * by count lines.  The size of the region is (region_bottom + 1 -
 * region_top) lines, because region_bottom is the row number of the
 * bottom line.
 */
void rectangle_scroll_up(const int top, const int left, const int bottom, const int right, const int count) {
	struct q_scrolline_struct * top_line;
	struct q_scrolline_struct * new_top_line;
	int remaining;
	int i;
	int j;

	if (top >= bottom) {
		return;
	}
	if (left >= right) {
		return;
	}

	/* Sanity check:  see if there will be any characters
	   left after the scroll */
	if (bottom + 1 - top <= count) {
		/* There won't be anything left in the region, so just
		   call erase_screen() and return. */
		erase_screen(top, left, bottom, right - 1, Q_FALSE);
		return;
	}

	/* Set new_top_line to the top of the scrolling region	*/
	new_top_line = find_top_scrollback_line();
	for (i = 0; i < top; i++) {
		if (new_top_line->next == NULL) {
			new_scrollback_line();
		}
		new_top_line = new_top_line->next;
	}

	/* Set top_line to the top line being scrolled up */
	top_line = new_top_line;
	remaining = bottom + 1 - top - count;
	for (i = 0; i < count; i++) {
		if (top_line->next == NULL) {
			new_scrollback_line();
		}
		top_line = top_line->next;
	}

	/* Copy the data between top_line and new_top_line */
	for (i = 0; i < remaining; i++) {
		if (new_top_line->length < top_line->length) {
			for (j = new_top_line->length; j < right; j++) {
				new_top_line->chars[j] = ' ';
				new_top_line->colors[j] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
				new_top_line->length = right;
			}
		}
		memcpy(new_top_line->chars + left, top_line->chars + left, sizeof(wchar_t) * (right - left));
		memcpy(new_top_line->colors + left, top_line->colors + left, sizeof(attr_t) * (right - left));
		new_top_line->length = top_line->length;
		new_top_line->double_width = top_line->double_width;
		new_top_line->double_height = top_line->double_height;
		new_top_line->dirty = Q_TRUE;
		new_top_line = new_top_line->next;
		top_line = top_line->next;

		if (top_line == NULL) {
			/*
			 * We are scrolling up on the very first screen and
			 * lines don't exist for the remainder of the scroll
			 * operation.  Just break out and let erase_screen()
			 * take care of things.
			 */
			break;
		}

	}

	erase_screen(top + remaining, left, bottom, right - 1, Q_FALSE);
} /* ---------------------------------------------------------------------- */

/*
 * Scroll the entire screen up count lines
 */
void scroll_up(const int count) {
	scrolling_region_scroll_up(0, HEIGHT - STATUS_HEIGHT - 1, count);
} /* ---------------------------------------------------------------------- */

void scrolling_region_scroll_up(const int region_top, const int region_bottom, const int count) {
	rectangle_scroll_up(region_top, 0, region_bottom, Q_MAX_LINE_LENGTH, count);
} /* ---------------------------------------------------------------------- */

/*
 * Scroll a region between region_top and region_bottom, inclusive,
 * down by count lines.	 The size of the region is (region_bottom + 1
 * - region_top) lines, because region_bottom is the row number of the
 * bottom line.
 */
void rectangle_scroll_down(const int top, const int left, const int bottom, const int right, const int count) {
	struct q_scrolline_struct * bottom_line;
	struct q_scrolline_struct * new_bottom_line;
	int remaining;
	int i;
	int j;

	if (top >= bottom) {
		return;
	}
	if (left >= right) {
		return;
	}

	/* Sanity check:  see if there will be any characters
	   left after the scroll */
	if (bottom + 1 - top <= count) {
		/* There won't be anything left in the region, so just
		   call erase_screen() and return. */
		erase_screen(top, left, bottom, right - 1, Q_FALSE);
		return;
	}

	/* Set new_bottom_line to the bottom of the scrolling region  */
	new_bottom_line = find_top_scrollback_line();
	for (i = 0; i < bottom; i++) {
		if (new_bottom_line->next == NULL) {
			new_scrollback_line();
		}
		new_bottom_line = new_bottom_line->next;
	}

	/* Set bottom_line to the bottom line being scrolled down */
	bottom_line = new_bottom_line;
	remaining = bottom + 1 - top - count;
	for (i = 0; i < count; i++) {
		if (bottom_line->prev == NULL) {
			/* We're trying to scroll down empty lines from the top line.
			   Insert a new blank line here. */
			insert_scrollback_line(bottom_line);
		}
		bottom_line = bottom_line->prev;
	}

	/* Copy the data between bottom_line and new_bottom_line */
	for (i = 0; i < remaining; i++) {
		if (new_bottom_line->length < left) {
			for (j = new_bottom_line->length; j < right; j++) {
				new_bottom_line->chars[j] = ' ';
				new_bottom_line->colors[j] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
				new_bottom_line->length = right;
			}
		}
		memcpy(new_bottom_line->chars + left, bottom_line->chars + left, sizeof(wchar_t) * (right - left));
		memcpy(new_bottom_line->colors + left, bottom_line->colors + left, sizeof(attr_t) * (right - left));
		new_bottom_line->length = bottom_line->length;
		new_bottom_line->double_width = bottom_line->double_width;
		new_bottom_line->double_height = bottom_line->double_height;
		new_bottom_line->dirty = Q_TRUE;
		new_bottom_line = new_bottom_line->prev;
		bottom_line = bottom_line->prev;
	}
	erase_screen(top, left, top + count - 1, right - 1, Q_FALSE);
} /* ---------------------------------------------------------------------- */

void scrolling_region_scroll_down(const int region_top, const int region_bottom, const int count) {
	rectangle_scroll_down(region_top, 0, region_bottom, Q_MAX_LINE_LENGTH, count);
} /* ---------------------------------------------------------------------- */

/*
 * Scroll the entire screen down count lines
 */
void scroll_down(const int count) {
	scrolling_region_scroll_down(0, HEIGHT - STATUS_HEIGHT - 1, count);
} /* ---------------------------------------------------------------------- */

/*
 * Move up one space
 */
void cursor_up(const int count, const Q_BOOL honor_scroll_region) {
	int i;
	int top;

	/* Special case: if a user moves the cursor from the right margin,
	   we have to reset the VT100 right margin flag. */
	if (count > 0) {
		vt100_wrap_line_flag = Q_FALSE;
	}

	for (i = 0; i < count; i++) {

		if (honor_scroll_region == Q_TRUE) {
			/* Honor the scrolling region */
			if (q_status.cursor_y < q_status.scroll_region_top) {
				/* Outside region, do nothing */
				return;
			}
			/* Inside region, go up */
			top = q_status.scroll_region_top;
		} else {
			/* Non-scrolling case */
			top = 0;
		}

		/* Non-scrolling case */
		if (q_status.cursor_y > top) {
			q_status.cursor_y--;
			q_scrollback_current = q_scrollback_current->prev;
		}
	} /* for (i = 0; i < count; i++) */
} /* ---------------------------------------------------------------------- */

/*
 * Move down one space
 */
void cursor_down(const int count, const Q_BOOL honor_scroll_region) {
	int i, j;
	int bottom;

	/* Special case: if a user moves the cursor from the right margin,
	   we have to reset the VT100 right margin flag. */
	if (count > 0) {
		vt100_wrap_line_flag = Q_FALSE;
	}

	for (i = 0; i < count; i++) {

		if (honor_scroll_region == Q_TRUE) {
			/* Honor the scrolling region */
			if (q_status.cursor_y > q_status.scroll_region_bottom) {
				/* Outside region, do nothing */
				return;
			}
			/* Inside region, go down */
			bottom = q_status.scroll_region_bottom;
		} else {
			/* Non-scrolling case */
			bottom = HEIGHT - STATUS_HEIGHT - 1;
		}

		if (q_status.cursor_y < bottom) {
			q_status.cursor_y++;
			if (q_scrollback_current->next == NULL) {
				new_scrollback_line();
				q_scrollback_current = q_scrollback_current->next;
				/* Pad spaces if necessary */
				for (j=q_status.cursor_x; j>q_scrollback_current->length;) {
					q_scrollback_last->colors[q_scrollback_last->length] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
					if (q_status.reverse_video == Q_TRUE) {
						q_scrollback_last->colors[q_scrollback_last->length] |= Q_A_REVERSE;
					}
					q_scrollback_last->length++;
				}
			} else {
				q_scrollback_current = q_scrollback_current->next;
			}
		}

		/* Capture */
		if (q_status.capture == Q_TRUE) {
			if (q_status.capture_type == Q_CAPTURE_TYPE_HTML) {
				/* HTML */
				fprintf(q_status.capture_file, "\n");
			} else if (q_status.capture_type == Q_CAPTURE_TYPE_NORMAL) {
				/* Normal */
				fprintf(q_status.capture_file, "\n");
			}
			fflush(q_status.capture_file);
			q_status.capture_flush_time = time(NULL);
			q_status.capture_x = 0;
		}

	} /* for (i = 0; i < count; i++) */
} /* ---------------------------------------------------------------------- */

/*
 * Move left
 */
void cursor_left(const int count, const Q_BOOL honor_scroll_region) {
	int i;

	/* Special case: if a user moves the cursor from the right margin,
	   we have to reset the VT100 right margin flag. */
	if (count > 0) {
		vt100_wrap_line_flag = Q_FALSE;
	}


	for (i = 0; i < count; i++) {
		if (honor_scroll_region == Q_TRUE) {
			/* Honor the scrolling region */
			if ((q_status.cursor_y < q_status.scroll_region_top) || (q_status.cursor_y > q_status.scroll_region_bottom)) {
				/* Outside region, do nothing */
				return;
			}
		}

		if (q_status.cursor_x > 0) {
			q_status.cursor_x--;
		}
	}
} /* ---------------------------------------------------------------------- */

/*
 * Move right
 */
void cursor_right(const int count, const Q_BOOL honor_scroll_region) {
	int i;
	int right_margin;

	/* Special case: if a user moves the cursor from the right margin,
	   we have to reset the VT100 right margin flag. */
	if (count > 0) {
		vt100_wrap_line_flag = Q_FALSE;
	}

	if (q_emulation_right_margin > 0) {
		right_margin = q_emulation_right_margin;
	} else {
		right_margin = WIDTH-1;
	}
	if (q_scrollback_current->double_width == Q_TRUE) {
		right_margin = ((right_margin + 1) / 2) - 1;
	}

	for (i = 0; i < count; i++) {
		if (honor_scroll_region == Q_TRUE) {
			/* Honor the scrolling region */
			if ((q_status.cursor_y < q_status.scroll_region_top) || (q_status.cursor_y > q_status.scroll_region_bottom)) {
				/* Outside region, do nothing */
				return;
			}
		}

		if (q_status.cursor_x < right_margin) {
			if (q_status.cursor_x >= q_scrollback_current->length) {
				/* Append a space and push the line out */
				q_scrollback_current->colors[q_scrollback_current->length] = q_current_color;
				q_scrollback_current->chars[q_scrollback_current->length] = ' ';
				q_scrollback_current->length++;
			}
			q_status.cursor_x++;
		}
	}
} /* ---------------------------------------------------------------------- */

/*
 * cursor_position
 */
void cursor_position(int row, int col) {
	assert(col >= 0);
	assert(row >= 0);
	int right_margin;

	if (q_emulation_right_margin > 0) {
		right_margin = q_emulation_right_margin;
	} else {
		right_margin = WIDTH-1;
	}
	if (q_scrollback_current->double_width == Q_TRUE) {
		right_margin = ((right_margin + 1) / 2) - 1;
	}

	/* Set column number */
	q_status.cursor_x = col;
	if (q_status.cursor_x > WIDTH - 1) {
		q_status.cursor_x = WIDTH - 1;
	}

	/* Sanity check, bring column back to margin. */
	if (q_emulation_right_margin > 0) {
		if (q_status.cursor_x > q_emulation_right_margin) {
			q_status.cursor_x = right_margin;
		}
	}

	/* Set row number */
	if (q_status.origin_mode == Q_TRUE) {
		row += q_status.scroll_region_top;
	}
	if (q_status.cursor_y < row) {
		cursor_down(row - q_status.cursor_y, Q_FALSE);
	} else if (q_status.cursor_y > row) {
		cursor_up(q_status.cursor_y - row, Q_FALSE);
	}

	vt100_wrap_line_flag = Q_FALSE;
} /* ---------------------------------------------------------------------- */

/*
 * Replace the characters in the current line from the start column (from 0
 * to WIDTH-1) to the end column (from 0 to WIDTH-1), inclusive.
 */
void fill_line_with_character(const int start, const int end, const wchar_t character, const Q_BOOL honor_protected) {
	int i;

	if (start > end) {
		return;
	}

	/* Mark line dirty */
	q_scrollback_current->dirty = Q_TRUE;

	if (q_scrollback_current->length < start) {
		for (i = q_scrollback_current->length; i < start; i++) {
			q_scrollback_current->chars[i] = ' ';
			q_scrollback_current->colors[i] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
			q_scrollback_current->length = start;
		}
	}

	/* Pad the characters leading up start */
	for (i = q_scrollback_current->length; i < start; i++) {
		q_scrollback_current->chars[i] = character;

		if (q_status.reverse_video == Q_TRUE) {

			switch (q_status.emulation) {
			case Q_EMUL_VT100:
			case Q_EMUL_VT102:
			case Q_EMUL_VT220:
				/*
				 * From the VT102 manual:
				 *
				 * Erasing a character also erases any character attribute of the
				 * character.
				 */
				q_scrollback_current->colors[i] = Q_A_REVERSE | scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
				break;
			default:
				/* Most other consoles erase with the current color a.k.a. back-color erase (bce). */
				q_scrollback_current->colors[i] = (q_current_color & ~Q_A_PROTECT);
				break;
			}

		} else {

			switch (q_status.emulation) {
			case Q_EMUL_VT100:
			case Q_EMUL_VT102:
			case Q_EMUL_VT220:
				/*
				 * From the VT102 manual:
				 *
				 * Erasing a character also erases any character attribute of the
				 * character.
				 */
				q_scrollback_current->colors[i] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
				break;
			default:
				/* Most other consoles erase with the current color a.k.a. back-color erase (bce). */
				q_scrollback_current->colors[i] = (q_current_color & ~Q_A_PROTECT);
				break;
			}
		}
	}

	/* Now erase from start to end */
	for (i = start; i <= end; i++) {
		if ((honor_protected == Q_FALSE) || ((honor_protected == Q_TRUE) && ((q_scrollback_current->colors[i] & Q_A_PROTECT) == 0))) {
			q_scrollback_current->chars[i] = character;

			if (q_status.reverse_video == Q_TRUE) {

				switch (q_status.emulation) {
				case Q_EMUL_VT100:
				case Q_EMUL_VT102:
				case Q_EMUL_VT220:
					/*
					 * From the VT102 manual:
					 *
					 * Erasing a character also erases any character attribute of the
					 * character.
					 */
					q_scrollback_current->colors[i] = Q_A_REVERSE | scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
					break;
				default:
					/* Most other consoles erase with the current color a.k.a. back-color erase (bce). */
					q_scrollback_current->colors[i] = (q_current_color & ~Q_A_PROTECT);
					break;
				}

			} else {

				switch (q_status.emulation) {
				case Q_EMUL_VT100:
				case Q_EMUL_VT102:
				case Q_EMUL_VT220:
					/*
					 * From the VT102 manual:
					 *
					 * Erasing a character also erases any character attribute of the
					 * character.
					 */
					q_scrollback_current->colors[i] = scrollback_full_attr(Q_COLOR_CONSOLE_TEXT);
					break;
				default:
					/* Most other consoles erase with the current color a.k.a. back-color erase (bce). */
					q_scrollback_current->colors[i] = (q_current_color & ~Q_A_PROTECT);
					break;
				}
			}

		}

	}

	/* If we erased beyond the end of the line, increase the line length
	   to include the new spaces. */
	if (end > q_scrollback_current->length) {
		q_scrollback_current->length = end + 1;
	}

	/* If the line is now longer than the screen, shorten it to WIDTH
	   or else we'll have problems with line wrapping. */
	if (q_scrollback_current->length > WIDTH) {
		q_scrollback_current->length = WIDTH;
	}
} /* ---------------------------------------------------------------------- */

/*
 * erase_line - erase the characters in the current line from the
 * start column (from 0 to WIDTH-1) to the end column (from 0 to
 * WIDTH-1), inclusive.
 */
void erase_line(const int start, const int end, const Q_BOOL honor_protected) {
	fill_line_with_character(start, end, ' ', honor_protected);
} /* ---------------------------------------------------------------------- */

/*
 * erase_screen - erase a rectangular section of the screen, inclusive.
 */
void erase_screen(const int start_row, const int start_col, const int end_row, const int end_col, const Q_BOOL honor_protected) {
	struct q_scrolline_struct * start_line;
	struct q_scrolline_struct * line;
	struct q_scrolline_struct * original_current_line;

	int i;

	if (
		(start_row < 0) || (start_col < 0) || (end_row < 0) || (end_col < 0) ||
		(end_row < start_row) || (end_col < start_col)
	) {
		return;
	}

	/* Hang onto the original cursor position */
	original_current_line = q_scrollback_current;

	/* Get to the starting line */
	start_line = find_top_scrollback_line();
	for (i = 0; i < start_row; i++) {
		if (start_line->next == NULL) {
			new_scrollback_line();
		}
		start_line = start_line->next;
	}

	line = start_line;

	for (i = start_row; i <= end_row; i++) {
		q_scrollback_current = line;
		erase_line(start_col, end_col, honor_protected);

		/* Erase display clears the double attributes */
		q_scrollback_current->double_width = Q_FALSE;
		q_scrollback_current->double_height = 0;

		/* Note: we don't add a line when (i == end_row)
		   because if end_row is the last line in scrollback
		   new_scrollback_line() will make the total screen
		   one line larger than it really is.  This causes
		   lots of trouble and looks like crap. */
		if ((line->next == NULL) && (i < end_row)) {
			/* Add a line */
			new_scrollback_line();
		}
		line = line->next;
	}

	/* Restore the cursor position */
	q_scrollback_current = original_current_line;

} /* ---------------------------------------------------------------------- */

/*
 * cursor_carriage_return
 */
void cursor_carriage_return() {
	/* Reset line */
	q_status.cursor_x = 0;

	if (q_status.line_feed_on_cr == Q_TRUE) {
		cursor_linefeed(Q_FALSE);
	}

	vt100_wrap_line_flag = Q_FALSE;

	/* Pass a carriage return to a script if we're running one */
	if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
		script_print_character(0x0D);
	}
} /* ---------------------------------------------------------------------- */

/*
 * cursor_formfeed
 */
void cursor_formfeed() {
	int i;

	/*
	 * Print the remaining number of linefeeds to clear the screen,
	 * then home the cursor.
	 */
	for (i = q_status.cursor_y; i <= 2 * (HEIGHT - STATUS_HEIGHT - 1); i++) {
		cursor_linefeed(Q_FALSE);
	}

	/* Erase the whole screen also, because a scroll region might be set. */
	erase_screen(0, 0, HEIGHT - STATUS_HEIGHT - 1, WIDTH - 1, Q_FALSE);

	/* Finally, home the cursor */
	cursor_position(0, 0);
} /* ---------------------------------------------------------------------- */

/*
 * cursor_linefeed
 */
void cursor_linefeed(const Q_BOOL new_line_mode) {
	int i;

	/* Capture */
	if (q_status.capture == Q_TRUE) {
		if (q_status.capture_type == Q_CAPTURE_TYPE_HTML) {
			/* HTML */
			fprintf(q_status.capture_file, "\n");
		} else if (q_status.capture_type == Q_CAPTURE_TYPE_NORMAL) {
			/* Normal */
			fprintf(q_status.capture_file, "\n");
		}
		if (q_status.capture_flush_time != time(NULL)) {
			fflush(q_status.capture_file);
			q_status.capture_flush_time = time(NULL);
		}
	}

	if (q_status.cursor_y < q_status.scroll_region_bottom) {
		/* Increment screen y */
		q_status.cursor_y++;

		/* New line */
		if (q_scrollback_current->next == NULL) {
			new_scrollback_line();
		}

		/* Write into the new line. */
		q_scrollback_current = q_scrollback_current->next;

	} else {
		/* Screen y does not increment */

		/* Two cases:  either we're inside a scrolling region
		   or not.  If the scrolling region bottom is the bottom
		   of the screen, then push the top line into the buffer.
		   Else scroll the scrolling region up. */
		if ((q_status.scroll_region_bottom == HEIGHT - STATUS_HEIGHT - 1) && (q_status.scroll_region_top == 0)) {

			/* We're at the bottom of the scroll region, AND the
			   scroll region is the entire screen. */

			/* New line */
			if (q_scrollback_current->next == NULL) {
				new_scrollback_line();
			}
			/* Write into the new line. */
			q_scrollback_current = q_scrollback_current->next;

			/* Set length to current X */
			q_scrollback_current->length = q_status.cursor_x;

			if (q_status.reverse_video == Q_TRUE) {
				/* Increase width and pad with spaces so that inverse video
				   appears everwhere. */
				q_scrollback_current->length = WIDTH;
			}

			/* Mark every line on the screen dirty */
			struct q_scrolline_struct * line = q_scrollback_position;
			for (i = 0; i < HEIGHT - STATUS_HEIGHT - 1; i++) {
				if (line != NULL) {
					line->dirty = Q_TRUE;
					line = line->prev;
				}
			}

		} else {
			/* We're at the bottom of the scroll region, AND the
			   scroll region is NOT the entire screen. */

			scrolling_region_scroll_up(q_status.scroll_region_top, q_status.scroll_region_bottom, 1);
		}
	}

	if (new_line_mode == Q_TRUE) {
		q_status.cursor_x = 0;
	}

	vt100_wrap_line_flag = Q_FALSE;

	/* Pass a carriage return to a script if we're running one */
	if (q_program_state == Q_STATE_SCRIPT_EXECUTE) {
		script_print_character(0x0A);
	}
} /* ---------------------------------------------------------------------- */

/*
 * delete_character
 */
void delete_character(const int count) {
	int i;

	for (i = 0; i < count; i++) {

		/* cursor_x and cursor_y don't change.	We just copy the existing line
		   leftwise one char. */
		memmove(&q_scrollback_current->chars[q_status.cursor_x],
			&q_scrollback_current->chars[q_status.cursor_x + 1],
			sizeof(wchar_t) * (Q_MAX_LINE_LENGTH - q_status.cursor_x - 1));
		memmove(&q_scrollback_current->colors[q_status.cursor_x],
			&q_scrollback_current->colors[q_status.cursor_x + 1],
			sizeof(attr_t) * (Q_MAX_LINE_LENGTH - q_status.cursor_x - 1));

		if (q_scrollback_current->length > q_status.cursor_x) {
			q_scrollback_current->length--;
		}
	}
	q_scrollback_current->dirty = Q_TRUE;
	return;
} /* ---------------------------------------------------------------------- */

/*
 * insert_blanks
 */
void insert_blanks(const int count) {
	int i;

	/* cursor_x and cursor_y don't change.	We just copy the existing line
	 rightwise one char. */
	memmove(&q_scrollback_current->chars[q_status.cursor_x + count],
		&q_scrollback_current->chars[q_status.cursor_x],
		sizeof(wchar_t) * (Q_MAX_LINE_LENGTH - q_status.cursor_x - count));
	memmove(&q_scrollback_current->colors[q_status.cursor_x + count],
		&q_scrollback_current->colors[q_status.cursor_x],
		sizeof(attr_t) * (Q_MAX_LINE_LENGTH - q_status.cursor_x - count));

	for (i = 0; i < count; i++) {
		q_scrollback_current->chars[q_status.cursor_x] = ' ';
		q_scrollback_current->colors[q_status.cursor_x] = q_current_color;
		if (q_scrollback_current->length < Q_MAX_LINE_LENGTH) {
			q_scrollback_current->length++;
		}

	}
	q_scrollback_current->dirty = Q_TRUE;
	return;
} /* ---------------------------------------------------------------------- */

void render_screen_to_debug_file(FILE * file) {
	struct q_scrolline_struct * line;
	int row;
	int renderable_lines;
	int i;

	fprintf(file, "\n");
	fprintf(file, "Variables:\n");
	fprintf(file, "	   HEIGHT: %d STATUS_HEIGHT: %d\n", HEIGHT, STATUS_HEIGHT);
	fprintf(file, "	   q_status.scrollback_lines: %d\n", q_status.scrollback_lines);
	fprintf(file, "	   q_scrollback_buffer:	  %p\n", q_scrollback_buffer);
	fprintf(file, "	   q_scrollback_last:	  %p\n", q_scrollback_last);
	fprintf(file, "	   q_scrollback_position: %p\n", q_scrollback_position);
	fprintf(file, "	   q_scrollback_current:  %p\n", q_scrollback_current);
	fprintf(file, "	   q_status.cursor_x: %d\n", q_status.cursor_x);
	fprintf(file, "	   q_status.cursor_y: %d\n", q_status.cursor_y);
	fprintf(file, "	   q_emulation_right_margin: %d\n", q_emulation_right_margin);
	fprintf(file, "	   vt100_wrap_line_flag:     %d\n", vt100_wrap_line_flag);
	fprintf(file, "	   q_status.reverse_video:	  %d\n", q_status.reverse_video);
	fprintf(file, "	   q_status.insert_mode:	  %d\n", q_status.insert_mode);
	fprintf(file, "	   q_status.scroll_region_top:	  %d\n", q_status.scroll_region_top);
	fprintf(file, "	   q_status.scroll_region_bottom: %d\n", q_status.scroll_region_bottom);
	fprintf(file, "\n");

	row = HEIGHT - 1;

	/* Skip the status line */
	if (q_status.status_visible == Q_TRUE) {
		row -= STATUS_HEIGHT;
	}

	/* Let's assert that row > 0.  Konsole and xterm won't
	   let the window size reach zero so this should be a
	   non-issue. */
	assert(row > 0);

	/* Count the lines available */
	renderable_lines = 0;
	line = q_scrollback_position;
	while (row >= 0) {
		renderable_lines++;
		if (line->prev == NULL) {
			break;
		}
		line = line->prev;
		row--;
	}

	if ((row<0) && (line->next != NULL)) {
		line = line->next;
	}

	fprintf(file, "----------------SCREEN BEGIN----------------\n");
	/* Now loop from line onward */
	for (row=0; row<renderable_lines; row++) {
		fprintf(file, "(%p) ", line);
		for (i = 0; i < line->length; i++) {

			if (line->double_width == Q_TRUE) {
				if ((2 * i) >= WIDTH) {
					break;
				}

				/* Print Q_A_PROTECT attribute */
				if (line->colors[i] & Q_A_PROTECT) {
					fprintf(file, "|");
				} else {
					fprintf(file, " ");
				}
				fprintf(file, "%lc", line->chars[i]);

				/* Print Q_A_PROTECT attribute */
				if (line->colors[i] & Q_A_PROTECT) {
					fprintf(file, "|");
				} else {
					fprintf(file, " ");
				}
				fprintf(file, " ");

			} else {

				if (i >= WIDTH) {
					break;
				}

				/* Print Q_A_PROTECT attribute */
				if (line->colors[i] & Q_A_PROTECT) {
					fprintf(file, "|");
				} else {
					fprintf(file, " ");
				}

				fprintf(file, "%lc", line->chars[i]);
			}
		}
		/* Clear remainder of line */
		fprintf(file, "\n");

		/* Point to next line */
		line = line->next;
	}

	for (row=renderable_lines; row<HEIGHT - STATUS_HEIGHT; row++) {
		fprintf(file, "(%p) <FILL LINE...>\n", (int *)0x0);
	}

	fprintf(file, "----------------SCREEN END------------------\n");

} /* ---------------------------------------------------------------------- */

/*
 * Flip the ACS_REVERSE toggle on every character in the visible
 * portion of the scrollback buffer.
 */
void invert_scrollback_colors() {
	struct q_scrolline_struct * original_current_line;
	int row;
	int bottom;
	int i;

	bottom = HEIGHT - STATUS_HEIGHT - 1;

	original_current_line = q_scrollback_current;
	q_scrollback_current = find_top_scrollback_line();
	for (row=0; row <= bottom; row++) {
		q_scrollback_current->dirty = Q_TRUE;
		if (q_scrollback_current->length < WIDTH) {
			erase_line(q_scrollback_current->length, WIDTH - 1, Q_FALSE);
		}
		for (i = 0; i < Q_MAX_LINE_LENGTH; i++) {
			if (i >= WIDTH) {
				break;
			}
			if ((q_scrollback_current->colors[i] & Q_A_REVERSE) != 0) {
				q_scrollback_current->colors[i] &= ~Q_A_REVERSE;
			} else {
				q_scrollback_current->colors[i] |= Q_A_REVERSE;
			}
		}
		if ((q_scrollback_current->next == NULL) && (row < bottom)) {
			new_scrollback_line();
		}

		/* Point to next line */
		q_scrollback_current = q_scrollback_current->next;
	}

	q_scrollback_current = original_current_line;
} /* ---------------------------------------------------------------------- */

/*
 * Flip the ACS_REVERSE toggle on every character in the visible
 * portion of the scrollback buffer.
 */
void deinvert_scrollback_colors() {
	/* These two functions are now identical */
	invert_scrollback_colors();
} /* ---------------------------------------------------------------------- */

/*
 * Set double_width for this line
 */
void set_double_width(Q_BOOL double_width) {
	q_scrollback_current->double_width = double_width;
} /* ---------------------------------------------------------------------- */

/*
 * Set double_height for this line
 */
void set_double_height(int double_height) {
	q_scrollback_current->double_height = double_height;
} /* ---------------------------------------------------------------------- */
